/*{ cm3_semaphore.c ***********************************************
 *
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported schedulers: FreeRTOS
}*/
#include "cm3_semaphore.h"

//{ functions *************************************************

// called when actively waiting for semaphore
void cm3_semaphore_wait() {
    ttc_task_yield();
}

                u32_t ttc__LDREXW(u32_t* Address) { // for debugging purpose
    Assert(ttc_memory_is_writable(Address), ec_InvalidArgument);
    return __LDREXW(Address);
}
     cm3_Semaphore_t* cm3_semaphore_create() {

    cm3_Semaphore_t* NewSemaphore = (cm3_Semaphore_t*) ttc_memory_alloc( sizeof(cm3_Semaphore_t) );
    Assert_SEMAPHORES(NewSemaphore, ec_Malloc);

    return NewSemaphore;
}
                 void cm3_semaphore_init(cm3_Semaphore_t* Semaphore) {
    Assert_SEMAPHORES( (((u32_t) Semaphore) & 3) == 0, ec_InvalidImplementation); // 4 byte alignment required for semaphores (e.g.: do not embedd into packed structures)!

    // initialization is pretty simple
    *Semaphore = 0;
}
ttc_semaphore_error_e cm3_semaphore_take(cm3_Semaphore_t* Semaphore, Base_t Amount, Base_t TimeOut) {
    Assert_SEMAPHORES(Semaphore != 0, ec_InvalidArgument);    

    if ( cm3_semaphore_try_take(Semaphore, Amount) == tsme_OK ) // general case first: try to obtain Amount
        return tsme_OK;                        // got Amount: simple and fast

    if      (TimeOut == -1) { // take with endless timeout
        // wait until Amount is free
        while (cm3_semaphore_try_take(Semaphore, Amount) != tsme_OK)
            cm3_semaphore_wait();

        return tsme_OK;
    }
    else if (TimeOut ==  0)   // no timeout: we already tried to obtain semaphore without timeout
        return tsme_SemaphoreExpired;
    else {                    // take within timeout
        Base_t CurrentTime = ttc_task_get_tick_count();
        Base_t EndTime     = CurrentTime + TimeOut;
        ttc_semaphore_error_e AmountFailed = tsme_SemaphoreExpired;

        while ( (EndTime < CurrentTime) && // tick-timer can overflow during wait: wait until overflow
                AmountFailed                 // or until successfull Amount
                ) {
            AmountFailed = cm3_semaphore_try_take(Semaphore, Amount);
            if (AmountFailed) {
                CurrentTime = ttc_task_get_tick_count();
                cm3_semaphore_wait();
            }
        }

        while ( AmountFailed &&
                (CurrentTime < EndTime)
              ) {         // loop until successfull Amount or timeout
            AmountFailed = cm3_semaphore_try_take(Semaphore, Amount);
            if (AmountFailed) {
                CurrentTime = ttc_task_get_tick_count();
                cm3_semaphore_wait();
            }
        };

        if (!AmountFailed)
            return tsme_OK;

        return tsme_TimeOut;
    }

    Assert_Halt_EC(ec_InvalidImplementation);
    return tsme_UNKNOWN; // this line should not be reachable
}
ttc_semaphore_error_e cm3_semaphore_try_take(cm3_Semaphore_t* Semaphore, Base_t Amount) {
    Assert_SEMAPHORES(Semaphore != 0, ec_InvalidArgument);
    Assert_SEMAPHORES( (((u32_t) Semaphore) & 3) == 0, ec_InvalidImplementation); // 4 byte alignment required for semaphores (e.g.: do not embedd into packed structures)!

    // ToDo: speed up operation by replacing function calls with single inline assembly
    // __ASM volatile ("ldrexb %0, [%1]" : "=r" (result) : "r" (addr) );
    // __ASM volatile ("strexb %0, %2, [%1]" : "=&r" (result) : "r" (addr), "r" (value) );
    // __ASM volatile ("clrex");

#if (CM3_SEMAPHORE_SIZE == 8) || (CM3_SEMAPHORE_SIZE == 16) || (CM3_SEMAPHORE_SIZE == 32)
    while (1) { // try to give back until read-write modification passed uninterrupted
#if CM3_SEMAPHORE_SIZE == 8 // byte sized semaphores
        u8_t TokensAvailable =__LDREXB(Semaphore);
        if (TokensAvailable < Amount)
            return tsme_SemaphoreExpired;
        TokensAvailable -= Amount;
        if (! __STREXB(TokensAvailable, Semaphore)) { // read-write modification was not interrupted
            __CLREX(); // remove lock created by __LDREXB()
            return tsme_OK;
        }
        __CLREX(); // remove lock created by __LDREXB()
#endif
#if CM3_SEMAPHORE_SIZE == 16 // half word sized semaphores
        u16_t TokensAvailable =__LDREXH(Semaphore);
        if (TokensAvailable < Amount)
            return tsme_SemaphoreExpired;
        TokensAvailable -= Amount;
        if (! __STREXH(TokensAvailable, Semaphore)) { // read-write modification was not interrupted
            __CLREX(); // remove lock created by __LDREXH()
            return tsme_OK;
        }
        __CLREX(); // remove lock created by __LDREXH()
#endif

#if CM3_SEMAPHORE_SIZE == 32 // word sized semaphores
        u32_t TokensAvailable = __LDREXW(Semaphore);
        if (TokensAvailable < Amount)
            return tsme_SemaphoreExpired;
        TokensAvailable -= Amount;
        if (! __STREXW(TokensAvailable, Semaphore)) { // read-write modification was not interrupted
            __CLREX(); // remove lock created by __LDREXW()
            return tsme_OK;
        }
        __CLREX(); // remove lock created by __LDREXW()
#endif
    }
#else
    Assert_Halt_EC(ec_NotImplemented); // missing implementation or wrong definition of CM3_SEMAPHORE_SIZE
    return tsme_NotImplemented;
#endif
}
ttc_semaphore_error_e cm3_semaphore_give(cm3_Semaphore_t* Semaphore, Base_t Amount) {
    Assert_SEMAPHORES(Semaphore != 0, ec_InvalidArgument);

    // ToDo: speed up operation by replacing function calls with single inline assembly
    // __ASM volatile ("ldrexb %0, [%1]" : "=r" (result) : "r" (addr) );
    // __ASM volatile ("strexb %0, %2, [%1]" : "=&r" (result) : "r" (addr), "r" (value) );
    // __ASM volatile ("clrex");

#if (CM3_SEMAPHORE_SIZE == 8) || (CM3_SEMAPHORE_SIZE == 16) || (CM3_SEMAPHORE_SIZE == 32)
#if CM3_SEMAPHORE_SIZE == 8 // byte sized semaphores
    cm3_atomic_add_8(Semaphore, (u8_t) Amount);
#endif
#if CM3_SEMAPHORE_SIZE == 16 // byte sized semaphores
    cm3_atomic_add_16Semaphore, (u16_t) Amount);
#endif
#if CM3_SEMAPHORE_SIZE == 32 // byte sized semaphores
    cm3_atomic_add_32(Semaphore, (u32_t) Amount);
#endif
    return tsme_OK;
#else
#   warn Missing implementation or wrong definition of CM3_SEMAPHORE_SIZE
    Assert_Halt_EC(ec_NotImplemented); // missing implementation or wrong definition of CM3_SEMAPHORE_SIZE
    return tsme_NotImplemented;
#endif
}

// private functions ********************************************

//}functions
