/*{ freertos_mutex.c ***********************************************
 *
 * Written by Gregor Rebel 2012-2013
 *
 * Architecture independent support for inter task communication & synchronization by use of
 * multitasking scheduler FreeRTOS.
 *
}*/
#include "freertos_mutex.h"

// constants copied from queue.c

/* Constants used with the cRxLock and xTxLock structure members. */
#define queueUNLOCKED					( ( signed portBASE_TYPE ) -1 )
#define queueLOCKED_UNMODIFIED			( ( signed portBASE_TYPE ) 0 )

#define queueERRONEOUS_UNBLOCK			( -1 )

/* For internal use only. */
#define	queueSEND_TO_BACK				( 0 )
#define	queueSEND_TO_FRONT				( 1 )

/* Effectively make a union out of the xQUEUE structure. */
#define pxMutexHolder					pcTail
#define uxQueueType						pcHead
#define uxRecursiveCallCount			pcReadFrom
#define queueQUEUE_IS_MUTEX				NULL

/* Semaphores do not actually store or copy data, so have an items size of
zero. */
#define queueSEMAPHORE_QUEUE_ITEM_LENGTH ( ( unsigned portBASE_TYPE ) 0 )
#define queueDONT_BLOCK					 ( ( portTickType ) 0U )
#define queueMUTEX_GIVE_BLOCK_TIME		 ( ( portTickType ) 0U )

/* These definitions *must* match those in queue.h. */
#define queueQUEUE_TYPE_BASE				( 0U )
#define queueQUEUE_TYPE_MUTEX 				( 1U )
#define queueQUEUE_TYPE_COUNTING_SEMAPHORE	( 2U )
#define queueQUEUE_TYPE_BINARY_SEMAPHORE	( 3U )
#define queueQUEUE_TYPE_RECURSIVE_MUTEX		( 4U )

//{ functions *************************************************

xQUEUE_Structure* freertos_mutex_create() {

    return xSemaphoreCreateMutex();
}
ttc_mutex_error_e freertos_mutex_lock(xQUEUE_Structure* Mutex, Base_t TimeOut) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

    if (TimeOut != -1) {
        portTickType Ticks2Wait = (portTickType) TimeOut / (1000000 / configTICK_RATE_HZ);

        if (xSemaphoreTake(Mutex, Ticks2Wait)) {
            return tme_OK;
        }
        else {
            return tme_TimeOut;
        }
    }
    else { // wait forever
        if (0)
        { while (xSemaphoreTake(Mutex, (portTickType) -1) == 0); }
        else { // easier to debug: will call ttc_task_yield() periodically while waiting
            Base_t StartTime = ttc_task_get_tick_count();
            while (xSemaphoreTake(Mutex, (portTickType) 1000) == 0) {
                _ttc_mutex_wait();
                Assert_MUTEXES(ttc_task_get_tick_count() - StartTime < 10000000, ec_TimeOut); // waiting >10 seconds (probably a bug?)
            }
        }
        return tme_OK;
    }

    return tme_NotImplemented;
}
ttc_mutex_error_e freertos_mutex_lock_isr(xQUEUE_Structure* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);
    static signed portBASE_TYPE HigherPrioTaskWoken; HigherPrioTaskWoken = FALSE;

#ifdef EXTENSION_300_scheduler_free_rtos
    signed portBASE_TYPE Return;

    Return = xSemaphoreTakeFromISR(Mutex, &HigherPrioTaskWoken);

    if (Return) {
        if (HigherPrioTaskWoken)
            return tme_WasLockedByTask;
        else
            return tme_OK;
    }
    else
        return tme_MutexAlreadyLocked;
#endif

    return tme_NotImplemented;
}
ttc_mutex_error_e freertos_mutex_unlock(xQUEUE_Structure* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    if ( xSemaphoreGive(Mutex) )
        return tme_OK;
    else
        return tme_UNKNOWN;
#endif

    return tme_NotImplemented;
}
ttc_mutex_error_e freertos_mutex_init(xQUEUE_Structure* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

    /* Information required for priority inheritance. */
    Mutex->pxMutexHolder = NULL;
    Mutex->uxQueueType = queueQUEUE_IS_MUTEX;

    /* Queues used as a mutex no data is actually copied into or out
    of the queue. */
    Mutex->pcWriteTo = NULL;
    Mutex->pcReadFrom = NULL;

    /* Each mutex has a length of 1 (like a binary semaphore) and
    an item size of 0 as nothing is actually copied into or out
    of the mutex. */
    Mutex->uxMessagesWaiting = ( unsigned portBASE_TYPE ) 0U;
    Mutex->uxLength = ( unsigned portBASE_TYPE ) 1U;
    Mutex->uxItemSize = ( unsigned portBASE_TYPE ) 0U;
    Mutex->xRxLock = queueUNLOCKED;
    Mutex->xTxLock = queueUNLOCKED;

#if ( configUSE_TRACE_FACILITY == 1 )
    Mutex->ucQueueType = queueQUEUE_TYPE_MUTEX;
#endif

    /* Ensure the event queues start with the correct state. */
    vListInitialise( &( Mutex->xTasksWaitingToSend ) );
    vListInitialise( &( Mutex->xTasksWaitingToReceive ) );

    traceCREATE_MUTEX( Mutex );

    /* Start with the semaphore in the expected state. */
    xQueueGenericSend( Mutex, NULL, ( portTickType ) 0U, queueSEND_TO_BACK );

    return tme_OK;
}
ttc_mutex_error_e freertos_mutex_unlock_isr(xQUEUE_Structure* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
    xSemaphoreGiveFromISR(Mutex, &xHigherPriorityTaskWoken);

    if (xHigherPriorityTaskWoken == pdFALSE)
        return tme_OK;
    else
        return tme_WasLockedByTask;
#endif

    return tme_NotImplemented;
}

//}functions
