/** { freertos_task.c ***********************************************
 *
 * Written by Gregor Rebel 2010-2013
 *
 * Low-level multitasking interface using FreeRTOS scheduler.
 *
}*/
#include "freertos_task.h"

//X // single linked list of extra info data for each created task
//X ft_extra_task_info_t* ExtraTaskInfos = NULL;

// provided by ttc_task.c
extern u8_t ttc_task_AmountTasks;

void vApplicationMallocFailedHook(void) {
    /* Called if a call to pvPortMalloc() fails because there is insufficient
       free memory available in the FreeRTOS heap.  pvPortMalloc() is called
       internally by FreeRTOS API functions that create tasks, queues, software
       timers, and semaphores.  The size of the FreeRTOS heap is set by the
       configTOTAL_HEAP_SIZE configuration constant in FreeRTOSConfig.h. */
    Assert(0, ec_Malloc);
}
void vApplicationStackOverflowHook( xTaskHandle pxTask, signed char *pcTaskName ) {
    ( void ) pcTaskName;
    ( void ) pxTask;

    /* Run time stack overflow checking is performed if
       configconfigCHECK_FOR_STACK_OVERFLOW is defined to 1 or 2.  This hook
       function is called if a stack overflow is detected. */
    Assert(0, ec_None);
}

#ifdef TTC_ASSERT_TASK
/** increased by _freertos_task_begin_criticalsection2(); decreased by freertos_task_end_criticalsection() */
u16_t freertos_task_CountCriticalSection = 0;
#endif

void freertos_task_usleep(Base_t Delay) {
    if (Delay >= DRIVER_ttc_task_min_delay) { // delay possible for current settings
        Delay /= DRIVER_ttc_task_min_delay;
        while (Delay > 0xffff) { // split long delays into smaller ones
            vTaskDelay(0xffff);
            Delay -= 0xffff;
        }
        if (Delay)
            vTaskDelay(Delay);
    }
}
BOOL freertos_task_create( void TaskFunction(void*), u16_t StackSize, void* Argument, u8_t Priority, ttc_task_info_t* TaskInfo) {
#if (configMAX_TASK_NAME_LEN < 4)
#error Value too small to store a pointer + a zero byte!
#endif
#if (configMAX_TASK_NAME_LEN > 4)
#warning Tasknames are managed by ttc_task. Using bigger configMAX_TASK_NAME_LEN than 5 will waste RAM!
#endif

    Assert_Task(TaskFunction, ec_InvalidArgument);
    Assert_Task(StackSize,    ec_InvalidArgument);
    Assert_Task(TaskInfo,     ec_InvalidArgument);

    if (Priority >= configMAX_PRIORITIES)
        Priority = configMAX_PRIORITIES - 1;

    TaskInfo->StackSize_Bytes = StackSize * sizeof(portSTACK_TYPE);

    portENTER_CRITICAL();
    ttc_task_AmountTasks++;
    portEXIT_CRITICAL();

    #define IMPLEMENTATION_freertos_task_create
    signed portBASE_TYPE Result = xTaskCreate( (pdTASK_CODE) TaskFunction,
                                               (const signed char *) NULL,
                                               (unsigned short ) StackSize,
                                               Argument,
                                               (Base_t) Priority,
                                               (void*) &(TaskInfo->Handle)
                                               );

    if (Result > 0) {
        volatile tskTCB* TCB =  (tskTCB*) TaskInfo->Handle;

        // storing task info pointer in task name for fast freertos_get_task_info()
        ttc_task_info_t** TCB_TaskInfo = (ttc_task_info_t**) &(TCB->pcTaskName);
        *TCB_TaskInfo = TaskInfo;

        return TRUE;
    }

    return FALSE;
}
ttc_task_info_t* freertos_get_task_info(TaskHandle_t Handle) {
    volatile tskTCB* TCB = NULL;

    if (Handle)
        TCB = (tskTCB*) Handle;
    else
        TCB = pxCurrentTCB;

    return (ttc_task_info_t*) TCB->pcTaskName; // set by freertos_task_create()
}
void freertos_update_task_info(TaskHandle_t Handle, ttc_task_info_t* Info) {
    Assert(Info, ec_InvalidArgument);
    volatile tskTCB* TCB = NULL;

    if (Handle)
        TCB = (tskTCB*) Handle;
    else
        TCB = pxCurrentTCB;

    Info->Handle     = (TaskHandle_t) TCB;
    Info->Priority   = TCB->uxBasePriority;
    Info->StackStart = (Base_t*) TCB->pxStack;
    Info->StackTop   = (Base_t*) TCB->pxTopOfStack;
#if INCLUDE_uxTaskGetStackHighWaterMark == 1
    Info->StackFree_HighWatermark = uxTaskGetStackHighWaterMark( (void*) TCB);
#else
    Info->StackFree_HighWatermark = 0;
#endif
#if (portSTACK_GROWTH < 0)
    Info->StackUsed = ((Base_t) Info->StackStart) + Info->StackSize_Bytes - ((Base_t) Info->StackTop);
#else
    Info->StackEnd   = (Base_t*) TCB->pxEndOfStack;
    Info->StackUsed =  ((Base_t) Info->StackTop) + Info->StackSize_Bytes - ((Base_t) Info->StackEnd);
    //D Info->StackUsed = Info->StackSize_Bytes - ( ((Base_t) Info->StackStart) - ((Base_t) Info->StackTop) );
#endif

    Info->StackUsed /= sizeof(portSTACK_TYPE);
}
ttc_task_info_t** _freertos_collect_task_infos(ttc_task_info_t**  TaskInfoWriter, xList *pxList, signed char cStatus) {
    volatile tskTCB *pxNextTCB, *pxFirstTCB;

    listGET_OWNER_OF_NEXT_ENTRY( pxFirstTCB, pxList );
    do
    {
        listGET_OWNER_OF_NEXT_ENTRY( pxNextTCB, pxList );
        ttc_task_info_t* TaskInfo = freertos_get_task_info( (tskTCB *) pxNextTCB);
        *TaskInfoWriter++ = TaskInfo;
        //X _freertos_update_task_info( (TaskHandle_t*) pxNextTCB, TaskInfoWriter++);
        freertos_update_task_info( (TaskHandle_t) pxNextTCB, TaskInfo++);
    } while( pxNextTCB != pxFirstTCB );

    // return pointer to next free info struct
    return TaskInfoWriter;
}
void freertos_task_get_all(ttc_task_all_t* TaskInfos) {
    Assert(TaskInfos, ec_InvalidArgument);
    ttc_task_info_t** TaskInfoWriter = &(TaskInfos->All[0]);
    TaskInfos->AmountAll              = 0;
    TaskInfos->AmountRunning          = 0;
    TaskInfos->AmountDelayed          = 0;
    TaskInfos->AmountOverflowDelayed  = 0;
    TaskInfos->AmountWaitTermination  = 0;
    TaskInfos->AmountSuspended        = 0;

    DRIVER_ttc_task_begin_criticalsection();

    unsigned portBASE_TYPE uxQueue;

    uxQueue = uxTopUsedPriority + ( unsigned portBASE_TYPE ) 1U;
    TaskInfos->Running = TaskInfos->All;
    do {                                                     // lists of all task priorities
        uxQueue--;

        if( listLIST_IS_EMPTY( &( pxReadyTasksLists[ uxQueue ] ) ) == pdFALSE ) {
            TaskInfoWriter = _freertos_collect_task_infos( TaskInfoWriter, ( xList * ) &( pxReadyTasksLists[ uxQueue ] ), tskREADY_CHAR );
        }
    } while( uxQueue > ( unsigned short ) tskIDLE_PRIORITY );
    TaskInfos->AmountRunning = TaskInfoWriter - TaskInfos->Running;
    if (!TaskInfos->AmountRunning)
        TaskInfos->Running = NULL;

    TaskInfos->Delayed = TaskInfoWriter;
    if( listLIST_IS_EMPTY( pxDelayedTaskList ) == pdFALSE ) { // list of delayed tasks
        TaskInfoWriter = _freertos_collect_task_infos( TaskInfoWriter, ( xList * ) pxDelayedTaskList, tskBLOCKED_CHAR );
    }
    TaskInfos->AmountDelayed = TaskInfoWriter - TaskInfos->Delayed;
    if (!TaskInfos->AmountDelayed)
        TaskInfos->Delayed = NULL;

    TaskInfos->OverflowDelayed = TaskInfoWriter;
    if( listLIST_IS_EMPTY( pxOverflowDelayedTaskList ) == pdFALSE ) {
        TaskInfoWriter = _freertos_collect_task_infos( TaskInfoWriter, ( xList * ) pxOverflowDelayedTaskList, tskBLOCKED_CHAR );
    }
    TaskInfos->AmountOverflowDelayed = TaskInfoWriter - TaskInfos->OverflowDelayed;
    if (!TaskInfos->AmountOverflowDelayed)
        TaskInfos->OverflowDelayed = NULL;

    TaskInfos->WaitTermination = TaskInfoWriter;
#if( INCLUDE_vTaskDelete == 1 )
{
        if( listLIST_IS_EMPTY( &xTasksWaitingTermination ) == pdFALSE ) {
            TaskInfoWriter = _freertos_collect_task_infos( TaskInfoWriter, &xTasksWaitingTermination, tskDELETED_CHAR );
        }
}
#endif
    TaskInfos->AmountWaitTermination = TaskInfoWriter - TaskInfos->WaitTermination;
    if (!TaskInfos->AmountWaitTermination)
        TaskInfos->WaitTermination = NULL;

    TaskInfos->Suspended = TaskInfoWriter;
#if ( INCLUDE_vTaskSuspend == 1 )
{
        if( listLIST_IS_EMPTY( &xSuspendedTaskList ) == pdFALSE )
        {
            TaskInfoWriter = _freertos_collect_task_infos( TaskInfoWriter, &xSuspendedTaskList, tskSUSPENDED_CHAR );
        }
}
#endif
    TaskInfos->AmountSuspended = TaskInfoWriter - TaskInfos->Suspended;
    if (!TaskInfos->AmountSuspended)
        TaskInfos->Suspended = NULL;

    DRIVER_ttc_task_end_criticalsection();
    TaskInfos->AmountAll = TaskInfoWriter - &(TaskInfos->All[0]);
}
