#ifndef freertos_task_H
#define freertos_task_H

/** { freertos_task.h ***********************************************
 *
 * Written by Gregor Rebel 2010-2013
 *
 * Low-level multitasking interface using FreeRTOS scheduler.
 * 
}*/

//{ includes

#include "ttc_basic.h"
#include "freertos_task_types.h"
#include "ttc_task_types.h"
//? #include <string.h>
//? #include "stddef.h"
#include "FreeRTOS.h"
#include "task.h"
#include "portmacro.h"

//}includes
//{ constants

#ifndef configTICK_RATE_HZ
#  error configTICK_RATE_HZ not defined, check your FreeRTOSConfig.h!
#  define configTICK_RATE_HZ 1
#endif

#ifndef freertos_task_MINIMAL_STACK_SIZE
#  ifdef configMINIMAL_STACK_SIZE
#    define freertos_task_MINIMAL_STACK_SIZE configMINIMAL_STACK_SIZE
#  else
#    define freertos_task_MINIMAL_STACK_SIZE 128
#  endif
#endif

//}

/* DEPRECATED
// task info data that is not stored by FreeRTOS
typedef struct ft_extra_task_info_s {
    struct ft_extra_task_info_s* Next;            // struct can build a single linked list
    TaskHandle_t                 Handle;          // unique task identifier
    Base_t                       StackSize_Bytes; // size of stack allocated for this task
} ft_extra_task_info_t;
*/

//{ Function prototypes **************************************************

#define DRIVER_ttc_task_begin_criticalsection() portENTER_CRITICAL()
#define DRIVER_ttc_task_end_criticalsection()   portEXIT_CRITICAL()
#if ( INCLUDE_vTaskSuspend == 1 )
#  define DRIVER_ttc_task_resume(Handle)          vTaskResume(Handle)
#  define DRIVER_ttc_task_resume_isr(Handle)      xTaskResumeFromISR(Handle)
#  define DRIVER_ttc_task_suspend(Handle)         vTaskSuspend(Handle)
#endif
#define DRIVER_ttc_task_start_scheduler()       vTaskStartScheduler()
#define DRIVER_ttc_task_yield_isr()             vPortYieldFromISR()
#define DRIVER_ttc_task_yield()                 taskYIELD()
#define DRIVER_ttc_task_create(TaskFunction,StackSize,Argument,Priority,Handle) freertos_task_create(TaskFunction,StackSize,Argument,Priority,Handle)
#define DRIVER_ttc_task_current_handle()        ((TaskHandle_t) pxCurrentTCB)
#define DRIVER_ttc_task_update_info(Handle, Info)      freertos_update_task_info(Handle, Info)
#define DRIVER_ttc_task_usleep(Delay)           freertos_task_usleep(Delay)
#define DRIVER_ttc_task_min_delay               (1000000 / configTICK_RATE_HZ) // smallest delay (usecs) depends on tick rate <-- test setting via ttc_task_calibrate_usleep()
#define DRIVER_ttc_task_get_amount()            uxTaskGetNumberOfTasks()
#define DRIVER_ttc_task_get_all(Buffer)         freertos_task_get_all(Buffer)
#define DRIVER_ttc_task_get_tick_count()        (1000000 / (configTICK_RATE_HZ)) * freertos_get_tick_count()


#if defined(INCLUDE_uxTaskGetStackHighWaterMark)
#  define DRIVER_ttc_task_check_stack()           uxTaskGetStackHighWaterMark(NULL)
#else
   // dummy value that is always > 0
#  define DRIVER_ttc_task_check_stack()           1
#endif

#if TARGET_DATA_WIDTH == 32 // direct access only allowed for 32-bit architectures
#  define freertos_get_tick_count()        xTickCount
#else
#  define freertos_get_tick_count()        xTaskGetTickCount()
#endif

/** calculate amount of system timer ticks required for given delay
 * @param Delay    delay time (usecs)
 * @return         amount of timer increments according to current system settings
 */
Base_t freertos_calculate_ticks(Base_t Delay);
#define freertos_calculate_ticks(MICROSECONDS)  MICROSECONDS * configTICK_RATE_HZ / 1000000


/** set task to sleep for some amount of time
 * @param Delay    delay time (usecs)
 */
void freertos_task_usleep(Base_t Delay);

/** spawns a function as separate task
 * @param TaskFunction  function to start as task (must never return!)
 * @param StackSize     each task has its own size; choose wisely
 * @param Argument      passed as argument to TaskFunction()
 * @param Priority      tasks with higher priority get more cpu time
 * @param TaskInfo      info structure where to store task handle (required to operate on task later
 * @return              == TRUE: task was created successfully
 */
BOOL freertos_task_create( void TaskFunction(void*), u16_t StackSize, void* Argument, u8_t Priority, ttc_task_info_t* TaskInfo);

/** returns pointer to configuration data of current/ given task
  *
  * Note: This function is very fast because it does not update configuration.
  *
  * @param Handle  !=NULL: get info of corresponding task
  *                ==NULL: get info of current task
  * @param Info    !=NULL: Info configuration of current task control block is copied into given variable
  *                ==NULL: a local static buffer will be allocated (do not use from multiple threads!)
  */
ttc_task_info_t* freertos_get_task_info(TaskHandle_t Handle);

/** collects configuration data of current task
  *
  * Note: This function is slow because it does update configuration data.
  *
  * @param Handle  !=NULL: get info of corresponding task
  *                ==NULL: get info of current task
  * @param Info    !=NULL: Info configuration of current task control block is copied into given variable
  *                ==NULL: a local static buffer will be allocated (do not use from multiple threads!)
  */
void freertos_update_task_info(TaskHandle_t Handle, ttc_task_info_t* Info);

/** spawns a function as separate task
 * @param  TaskInfos reference of array where to store task-infos (use DRIVER_ttc_task_get_amount() to calculate size!)
 * @return TaskInfos is loaded with data of all tasks (Make sure memory block big egnough!)
 */
void freertos_task_get_all(ttc_task_all_t* TaskInfos);

/** called in case pvPortMalloc() fails */
void vApplicationMallocFailedHook(void);

void vApplicationStackOverflowHook( xTaskHandle pxTask, signed char *pcTaskName );

//}Function prototypes

#endif
