#ifndef freertos_task_types_H
#define freertos_task_types_H

/** { freertos_task_types.h ***********************************************
 *
 * Written by Gregor Rebel 2010-2013
 *
 * Datatypes used by Low-level multitasking interface using FreeRTOS scheduler.
 * 
}*/

//{ includes

#include "ttc_basic.h"
#include "FreeRTOS.h"
#include "task.h"
#include "portmacro.h"

//}
//{ external datatypes ***********************************************

/*
 * Task control block.  A task control block (TCB) is allocated for each task,
 * and stores task state information, including a pointer to the task's context
 * (the task's run time environment, including register values)
 */
typedef struct tskTaskControlBlock
{
    volatile portSTACK_TYPE	*pxTopOfStack;		/*< Points to the location of the last item placed on the tasks stack.  THIS MUST BE THE FIRST MEMBER OF THE TCB STRUCT. */

    #if ( portUSING_MPU_WRAPPERS == 1 )
        xMPU_SETTINGS xMPUSettings;				/*< The MPU settings are defined as part of the port layer.  THIS MUST BE THE SECOND MEMBER OF THE TCB STRUCT. */
    #endif

    xListItem				xGenericListItem;		/*< The list that the state list item of a task is reference from denotes the state of that task (Ready, Blocked, Suspended ). */
    xListItem				xEventListItem;		/*< Used to reference a task from an event list. */
    unsigned portBASE_TYPE	uxPriority;			/*< The priority of the task.  0 is the lowest priority. */
    portSTACK_TYPE			*pxStack;			/*< Points to the start of the stack. */
    signed char				pcTaskName[ configMAX_TASK_NAME_LEN ];/*< Descriptive name given to the task when created.  Facilitates debugging only. */

    #if ( portSTACK_GROWTH > 0 )
        portSTACK_TYPE *pxEndOfStack;			/*< Points to the end of the stack on architectures where the stack grows up from low memory. */
    #endif

    #if ( portCRITICAL_NESTING_IN_TCB == 1 )
        unsigned portBASE_TYPE uxCriticalNesting; /*< Holds the critical section nesting depth for ports that do not maintain their own count in the port layer. */
    #endif

    #if ( configUSE_TRACE_FACILITY == 1 )
        unsigned portBASE_TYPE	uxTCBNumber;	/*< Stores a number that increments each time a TCB is created.  It allows debuggers to determine when a task has been deleted and then recreated. */
        unsigned portBASE_TYPE  uxTaskNumber;	/*< Stores a number specifically for use by third party trace code. */
    #endif

    #if ( configUSE_MUTEXES == 1 )
        unsigned portBASE_TYPE uxBasePriority;	/*< The priority last assigned to the task - used by the priority inheritance mechanism. */
    #endif

    #if ( configUSE_APPLICATION_TASK_TAG == 1 )
        pdTASK_HOOK_CODE pxTaskTag;
    #endif

    #if ( configGENERATE_RUN_TIME_STATS == 1 )
        unsigned long ulRunTimeCounter;			/*< Stores the amount of time the task has spent in the Running state. */
    #endif

} tskTCB;

#define TaskHandle_t tskTCB*

// defined in tasks.c
void prvListTaskWithinSingleList( const signed char *pcWriteBuffer, xList *pxList, signed char cStatus ) PRIVILEGED_FUNCTION;
extern volatile tskTCB *pxCurrentTCB;
extern xList pxReadyTasksLists[configMAX_PRIORITIES];
extern xList* pxDelayedTaskList;
extern xList* pxOverflowDelayedTaskList;
extern xList xPendingReadyList;
extern xList xTasksWaitingTermination;
extern xList xSuspendedTaskList;
extern unsigned portBASE_TYPE uxTopUsedPriority;
#define tskBLOCKED_CHAR		( ( signed char ) 'B' )
#define tskREADY_CHAR		( ( signed char ) 'R' )
#define tskDELETED_CHAR		( ( signed char ) 'D' )
#define tskSUSPENDED_CHAR	( ( signed char ) 'S' )


// System Tick Counter provided by FreeRTOS
extern volatile portTickType xTickCount;

//}external datatypes

#endif
