/** { ttc_can.h *******************************************************

                           The ToolChain
                           
   High-Level interface for CAN device.

   Structures, Enums and Defines being required by high-level can and application.

   Authors:
   
}*/

#ifndef TTC_CAN_H
#define TTC_CAN_H

//{ Includes *************************************************************

#include "ttc_basic.h"
#include "ttc_memory.h"

#ifdef EXTENSION_450_can_stm32f1
#  include "stm32f1/stm32f1_can.h"
#endif

#include "ttc_can_types.h"

//} Includes
//{ Check for missing low-level implementations **************************

// Note: functions declared below must be redefined to implementations provided by low-level Driver!

/** Prepares can Driver for basic operation.
 *
 * Note: This function is called automatically at system start before the scheduler has been started.
 */
void ttc_can_prepare();

/** loads configuration of indexed CAN interface with default values
 * @param Config        pointer to struct ttc_can_config_t (must have valid value for LogicalIndex)
 * @return  == 0:       configuration was loaded successfully
 */
ttc_can_errorcode_e Driver_can_load_defaults(ttc_can_config_t* Config);
#ifndef Driver_can_load_defaults
#  error: missing low-level implementation!
#endif

/** shutdown single CAN device
 * @param Config        pointer to struct ttc_can_config_t (must have valid value for LogicalIndex)
 * @return              == 0: CAN has been shutdown successfully; != 0: error-code
 */
ttc_can_errorcode_e Driver_can_deinit(ttc_can_config_t* Config);
#ifndef Driver_can_deinit
#  error: missing low-level implementation!
#endif

/** initializes single CAN
 * @param Config        pointer to struct ttc_can_config_t (must have valid value for LogicalIndex)
 * @return              == 0: CAN has been initialized successfully; != 0: error-code
 */
ttc_can_errorcode_e Driver_can_init(ttc_can_config_t* Config);
#ifndef Driver_can_init
#  error: missing low-level implementation!
#endif

//}
//{ Defines/ TypeDefs ****************************************************

//} Defines
//{ Structures/ Enums ****************************************************
// Structures and Enums required by High-Level can only go here...

//} Structures/ Enums
//{ Function prototypes **************************************************

/** returns amount of devices available in current configuration
 *
 * @return amount of available devices
 */
u8_t ttc_can_get_max_index();

/** returns configuration of indexed device (asserts if no valid configuration was found)
 *
 * @param LogicalIndex    index of device to init (1..ttc_CAN_get_max_LogicalIndex())
 * @return                configuration of indexed device
 */
ttc_can_config_t* ttc_can_get_configuration(u8_t LogicalIndex);

/** initialize indexed device for use
 *
 * @param LogicalIndex    device index of device to init (1..ttc_CAN_get_max_LogicalIndex())
 */
void ttc_can_init(u8_t LogicalIndex);

/** shutdown indexed device
 *
 * @param LogicalIndex    device index of device to shutdown (1..ttc_CAN_get_max_LogicalIndex())
 */
void ttc_can_deinit(u8_t LogicalIndex);

/** reset configuration of indexed device back to default
 *
 * @param LogicalIndex    device index of device to init (1..ttc_CAN_get_max_LogicalIndex())
 */
void ttc_can_reset(u8_t LogicalIndex);

// additional prototypes go here...
/** checks if indexed CAN bus already has been initialized
 * @param CAN_Index     device index of CAN to check (1..stm32_getMax_CAN_Index())
 * @return              == TRUE: indexed CAN unit is already initialized
 */
bool ttc_can_check_initialized(ttc_can_config_t* CAN_Generic);

/** fills out given CAN_Generic with maximum valid values for indexed CAN
 * @param CAN_Index     device index of CAN to init (1..ttc_can_get_max_index())
 * @param CAN_Generic   will be loaded with pointer to features of can
 * @return  == 0:       *CAN_Generic has been initialized successfully; != 0: error-code
 */
ttc_can_errorcode_e ttc_can_get_features(ttc_can_config_t* CAN_Generic);

/** loads filter configuration with default values
 * @param FilterConfig  will be loaded with pointer to features of can filter
 * @return  == 0:       configuration was loaded successfully
 */
ttc_can_errorcode_e ttc_can_filter_load_defaults(ttc_can_filter_config_t* FilterConfig);

/** initializes single CAN filter with its current configuration
 * @param FilterConfig  will be loaded with pointer to features of can filter
 * @return              == 0: CAN has been initialized successfully; != 0: error-code
 */
ttc_can_errorcode_e ttc_can_filter_init(ttc_can_filter_config_t* FilterConfig);

/** Send out given amount of raw bytes from buffer.
 * Will not interpret bytes in buffer.
 *
 * Note: CAN must be initialized before!
 * Note: This function is partially thread safe (use it for each CAN from one thread only)!
 * Note: This function blocks until all bytes have been sent!
 *
 * @param CAN_Generic   will be loaded with pointer to features of can
 * @param Buffer         memory location from which to read data
 * @param Amount         amount of bytes to send from Buffer[]
 * @return               == 0: Buffer has been sent successfully; != 0: error-code
 */
ttc_can_errorcode_e ttc_can_send_raw(ttc_can_config_t* CAN_Generic, const u8_t* Buffer, u16_t Amount);

/** Send out given bytes from buffer until first zero byte or MaxLength reached.
 *
 * Note: CAN must be initialized before!
 * Note: This function is partially thread safe (use it for each CAN from one thread only)!
 * Note: This function blocks until all bytes have been sent!
 *
 * @param CAN_Generic   will be loaded with pointer to features of can
 * @param Buffer         memory location from which to read data
 * @param MaxLength      no more than this amount of bytes will be send
 * @return               == 0: Buffer has been sent successfully; != 0: error-code
 */
ttc_can_errorcode_e ttc_can_send_string(ttc_can_config_t* CAN_Generic, const char* Buffer, u16_t MaxLength);

/** Send out given data word (8 or 9 bits)
 *
 * Note: CAN must be initialized before!
 * Note: This function is partially thread safe (use it for each CAN from one thread only)!
 * Note: This function blocks until all bytes have been sent!
 *
 * @param CAN_Generic   will be loaded with pointer to features of can
 * @param Byte           8 or 9 bit value to send
 * @return               == 0: Buffer has been sent successfully; != 0: error-code
 */
ttc_can_errorcode_e ttc_can_send_word(ttc_can_config_t* CAN_Generic, const u16_t Word);

/** Reads single data word from input buffer (8 or 9 bits)
 *
 * Note: CAN must be initialized before!
 * Note: This function is partially thread safe (use it for each CAN from one thread only)!
 * Note: This function blocks until a valid word has been received or timeout occurs
 *
 * @param CAN_Generic   will be loaded with pointer to features of can
 * @param Word           pointer to 16 bit buffer where to store Word
 * @param TimeOut        >0: max microseconds to wait; ==0: no timeout
 * @return               == 0: Word has been read successfully; != 0: error-code
 */
ttc_can_errorcode_e ttc_can_read_word(ttc_can_config_t* CAN_Generic, u16_t* Word);

/** Reads single data byte from input buffer.
 *
 * Note: CAN must be initialized before!
 * Note: This function is partially thread safe (use it for each CAN from one thread only)!
 * Note: This function blocks until a valid byte has been received or timeout occurs
 *
 * @param CAN_Generic   will be loaded with pointer to features of can
 * @param Word           pointer to 16 bit buffer where to store Word
 * @param TimeOut        >0: max microseconds to wait; ==0: no timeout
 * @return               == 0: Word has been read successfully; != 0: error-code
 */
ttc_can_errorcode_e ttc_can_read_byte(ttc_can_config_t* CAN_Generic, u8_t* Byte);

//} Function prototypes
//{ private Function prototypes ******************************************

/** 
 *
 * Note: This function is marked private and should be called from outside!
 *
 * @param 
 * @return
 */
void _ttc_can_();


//}private functions
//{ Macros ***************************************************************

//}Macros

#endif //TTC_CAN_H
