
#ifndef TTC_INTERRUPT_H
#define TTC_INTERRUPT_H

/*{ ttc_interrupt.h **********************************************************

                      The ToolChain

   Device independent support for interrupt requests

   Currently implemented architectures: stm32

   written by Gregor Rebel 2012


   How it works

   ttc_interrupt.h defines and documents an architecture independent interface to
   interrupt handling. It also checks which low-level features are provided by
   current low-level driver.
   Each low-level driver must declare a set of DRIVER_ttc_*() macros to register
   all implemented features.

   Different architecture drivers for usart, spi and others may define certain
   constants to add entries to interrupt definition structures.
   See ttc_interrupt_types.h for a list of valid constants.

}*/
//{ Defines/ TypeDefs ****************************************************


//} Defines
/* Defines/ TypeDefs/ structures being set by architecture driver
 *
 * Constant defines
 *
 * Structures
 *
 */
//{ Includes *************************************************************

#include "ttc_basic.h"
#include "ttc_gpio.h"
#include "ttc_interrupt_types.h"

// supported architectures
#ifdef TARGET_ARCHITECTURE_STM32
#  include "stm32/stm32_interrupt.h"
#endif
#ifndef physical_index_t
#  warning physical_index_t not defined!
#  define physical_index_t u8_t
#endif

//} Includes
//{ Structures/ Enums ****************************************************

//} Structures/ Enums
//{ Global Variables *****************************************************


//} Global Variables
//{ Function prototypes **************************************************

/** initializes interrupt for given Type to call given ISR
 * Note: This will not activate your interrupt. Call ttc_interrupt_usart_enable() afterwards to enable/ disable this interrupt.
 *
 * @param Type        interrupt Type to use
 * @param PhysicalIndex real physical index of functional unit (e.g. 0 = USART1, 1 = USART2, ...)
 * @param ISR           function pointer of interrupt service routine to call;
 *                      ISR() gets two params: u8_t Index, void* Reference
 *                      Index     index functional unit in microcontroller (e.g. 0=first USART)
 *                      Reference hardware dependent data used by low level driver (e.g. register base address)
 * @param Argument      will be passed as argument to ISR()
 * @param OldISR        != NULL: previous stored pointer will be saved into given address
 * @param OldArgument   != NULL: previous stored argument will be saved into given address
 */
ttc_interrupt_errorcode_e ttc_interrupt_init(ttc_interrupt_type_e Type, u8_t PhysicalIndex, void (*ISR)(u8_t, void*), void* Argument, void (**OldISR)(physical_index_t, void*), void** OldArgument);
#ifdef DRIVER_ttc_interrupt_init
#  define ttc_interrupt_init(Type, PhysicalIndex, ISR, Argument, OldISR, OldArgument) DRIVER_ttc_interrupt_init(Type, PhysicalIndex, ISR, Argument, OldISR, OldArgument)
#else
#  define ttc_interrupt_init(Type, PhysicalIndex)
#  warning Missing implementation for ttc_interrupt_init()
#endif

/** un-initialize given interrupt Type on given functional unit
 * @param Type     interrupt Type to use
 * @param PhysicalIndex  real physical index of functional unit (e.g. 0 = USART1, 1 = USART2, ...)
 */
ttc_interrupt_errorcode_e ttc_interrupt_deinit(ttc_interrupt_type_e Type, u8_t PhysicalIndex);
#define ttc_interrupt_deinit(Type, PhysicalIndex) ttc_interrupt_init(Type, PhysicalIndex, NULL, NULL, NULL, NULL)

/** activates/ deactivates a previously initialized interrupt
 * @param Type        interrupt Type to use
 * @param PhysicalIndex real physical index of functional unit (e.g. 0 = USART1, 1 = USART2, ...)
 * @param EnableIRQ     =0: disable interrupt; !=0: enable interrupt
 */
void ttc_interrupt_enable(ttc_interrupt_type_e Type, u8_t PhysicalIndex, BOOL EnableIRQ);
#ifdef DRIVER_ttc_interrupt_enable
#  define ttc_interrupt_enable(Type, PhysicalIndex, EnableIRQ) DRIVER_ttc_interrupt_enable(Type, PhysicalIndex, EnableIRQ)
#else
#  define ttc_interrupt_enable(Type, PhysicalIndex, EnableIRQ)
#  warning Missing implementation for ttc_interrupt_enable()
#endif

/** globally enables all interrupts
  */
void ttc_interrupt_all_enable();
#ifdef DRIVER_ttc_interrupt_all_enable
#define ttc_interrupt_all_enable() DRIVER_ttc_interrupt_all_enable()
#else
#define ttc_interrupt_all_enable()
#warning Missing implementation for DRIVER_ttc_interrupt_all_enable()
#endif

/** globally disables all interrupts
  */
void ttc_interrupt_all_disable();
#ifdef DRIVER_ttc_interrupt_all_disable
#define ttc_interrupt_all_disable() DRIVER_ttc_interrupt_all_disable()
#else
#define ttc_interrupt_all_disable()
#warning Missing implementation for DRIVER_ttc_interrupt_all_disable()
#endif

//} Function prototypes

#endif //TTC_INTERRUPT_H
