/** ttc_list.c ***********************************************{
 *
 * Written by Gregor Rebel 2013
 *
 * Architecture independent support for single linked lists.
 * 
 * Why to use lists instead of queues?
 * - lists have no size limit (every item brings its own next-pointer)
 * - lists require only as much memory as required (no oversized pre-allocation like queues)
 * - lists can be used from interrupt service routines too
 * - every pool-allocated memory block can be put into a list (-> ttc_memory.h)
 *
 * When not to prefer lists over queues?
 * - if you want to store big amount of very small blocks like single bytes
 *   (every item requires a next-pointer)
 *
}*/
#include "ttc_list.h"

//{ declare private functions (no locking/ argument checking) *

#ifdef TTC_REGRESSION_DISABLED
#define SIZE_LIST_HISTORY 70
#else
#define SIZE_LIST_HISTORY 0
#endif
#if SIZE_LIST_HISTORY > 0

  /** rotating array storing last processed lists */
  ttc_list_t ListHistory[SIZE_LIST_HISTORY];
  ttc_list_t* CurrentListHistory = ListHistory;

  void _ttc_list_add_history(ttc_list_t* List) {
      if (CurrentListHistory >= ListHistory+SIZE_LIST_HISTORY)
          CurrentListHistory = ListHistory;
      ttc_memory_copy(CurrentListHistory++, List, sizeof(ttc_list_t));
  }
#define ttc_list_add_history(LIST) _ttc_list_add_history(LIST)
#else
#define ttc_list_add_history(LIST)
#endif

/** completes all pending  on given list
 *
 * Note: This function is private and must not be called from outside!
 * Note: This function completes in constant time O(1)
 *
 * @param List     pointer to already created list
 */
void _ttc_list_complete_all_pending_isr(ttc_list_t* List);

/** adds given Item or linked list of items at end of given list
 *
 * Note: This function is private and must not be called from outside!
 * Note: This function completes in constant time O(1)
 *
 * @param List     pointer to already created list
 * @param Item     pointer to item to be added to list
 * @return         == tle_OK: Item has been successfully added to List
 */
void _ttc_list_push_back_single(ttc_list_t* List, ttc_list_item_t* Item);

/** adds given linked list of items at end of given list
 *
 * Note: This function is private and must not be called from outside!
 * Note: This function completes in constant time O(1)
 *
 * @param List       pointer to already created list
 * @param FirstItem  pointer to first item in list of items to be added to list
 * @param LastItem   pointer to last item in list of items to be added to list
 * @param Amount     Amount of items to be added to list
 */
void _ttc_list_push_back_multiple(ttc_list_t* List, ttc_list_item_t* FirstItem, ttc_list_item_t* LastItem, SmallBase_t Amount);

/** removes first Item from given list and delivers its reference
 *
 * Note: This function is private and must not be called from outside!
 * Note: This function completes in constant time O(1)
 *
 * @param List     pointer to already created list
 * @param Item     loaded with pointer to item removed from given list or NULL if list is empty
 */
void _ttc_list_pop_front_single(ttc_list_t* List, ttc_list_item_t** Item);

/** removes linked list of first items from front of given list and delivers reference to first and last one
 *
 * Note: This function is private and must not be called from outside!
 * Note: This function completes in constant time O(Amount)
 *
 * @param List      pointer to already created list
 * @param Amount    amount of items to remove from list (==-1: means all items in list)
 * @param FirstItem loaded with pointer to first item of items removed from given list or NULL if list is empty
 * @param LastItem  loaded with pointer to last item of items removed from given list or NULL if list is empty
 * @return          actual amount of items being removed from list
 */
SmallBase_t _ttc_list_pop_front_multiple(ttc_list_t* List, SmallBase_t Amount, ttc_list_item_t** FirstItem, ttc_list_item_t** LastItem);

//} private functions
//{ functions *************************************************

          Base_t ttc_list_count_items(ttc_list_item_t* FirstItem) {
    Base_t Counter = 0;
    ttc_list_item_t* CurrentItem = FirstItem;

    while (CurrentItem &&  (CurrentItem!=TTC_LIST_TERMINATOR) ) {
        Assert_LIST_WRITABLE(CurrentItem);
        CurrentItem = CurrentItem->Next;
        Counter++;
    }

    return Counter;
}
     ttc_list_t* ttc_list_create(const char* Name) {
    ttc_list_t* NewList = ttc_memory_alloc(sizeof(ttc_list_t));
    Assert_LIST(NewList != NULL, ec_Malloc); // could not create list (no implementation/ not egnough memory)

    ttc_list_init(NewList, Name);

    return NewList;
}
            void ttc_list_init(ttc_list_t* List, const char* Name) {
    Assert_LIST(List != NULL, ec_InvalidArgument);
    Assert_LIST(ttc_memory_is_writable( (void*) List ), ec_InvalidArgument);

    ttc_memory_set( (void*) List, 0, sizeof(List) );
    ttc_mutex_init( &(List->Lock) );
    ttc_semaphore_init( &(List->AmountItems) );
#ifdef TTC_REGRESSION
    List->Stats.Name = Name;
#endif
}
ttc_list_error_e ttc_list_complete_all_pending_isr(ttc_list_t* List) {
    Assert_LIST(List, ec_InvalidArgument);

    if (ttc_mutex_is_locked(&(List->Lock)) )
        return tle_ListInUseByTask;

    _ttc_list_complete_all_pending_isr(List);
    return tle_OK;
}
            void ttc_list_push_back_single(ttc_list_t* List, ttc_list_item_t* Item) {
    Assert_LIST(List, ec_InvalidArgument);
    Assert_LIST(Item, ec_InvalidArgument);

    ttc_mutex_lock(&(List->Lock), -1, ttc_list_push_back_single);

    _ttc_list_push_back_single(List, Item);
#ifdef TTC_REGRESSION
    Assert_LIST_WRITABLE(Item->Next);
#endif

    ttc_mutex_unlock(&(List->Lock));
    ttc_semaphore_give(&(List->AmountItems), 1);
}
            void ttc_list_push_back_multiple(ttc_list_t* List, ttc_list_item_t* FirstItem, ttc_list_item_t* LastItem, SmallBase_t Amount) {
     Assert_LIST(List,      ec_InvalidArgument);
     Assert_LIST(FirstItem, ec_InvalidArgument);
     Assert_LIST(LastItem,  ec_InvalidArgument);

     ttc_mutex_lock(&(List->Lock), -1, ttc_list_push_back_multiple);

     _ttc_list_push_back_multiple(List, FirstItem, LastItem, Amount);
     ttc_semaphore_give(&(List->AmountItems), Amount);

     Assert_LIST_WRITABLE(List->First);
     Assert_LIST_WRITABLE(List->Last);

     ttc_mutex_unlock(&(List->Lock));
}
ttc_list_error_e ttc_list_push_back_single_isr(ttc_list_t* List, ttc_list_item_t* Item) {
    Assert_LIST(List, ec_InvalidArgument);

    if (! ttc_mutex_is_locked(&(List->Lock)) ) { // not locked by task: add items to list
        if (List->Pending_Push_Amount)           // still data pending from last isr-push
            _ttc_list_complete_all_pending_isr(List);
        _ttc_list_push_back_single(List, Item);
        ttc_semaphore_give(&(List->AmountItems), 1);
    }
    else {                                       // already locked by task: add to list of pending items
#ifdef TTC_REGRESSION
        LIST_STATS_INCREASE(List, AmountPending);
#endif
        if (List->Pending_Push_Last) { // append at end of list
            List->Pending_Push_Last->Next = Item;
        }
        else {                         // append as first item
            List->Pending_Push_First = Item;
        }
        while (Item->Next)
            Item = Item->Next;
        List->Pending_Push_Last = Item;
        List->Pending_Push_Amount++;

        return tle_ListInUseByTask;
    }

    return tle_OK;
}
ttc_list_error_e ttc_list_push_back_multiple_isr(ttc_list_t* List, ttc_list_item_t* FirstItem, ttc_list_item_t* LastItem, SmallBase_t Amount) {
    Assert_LIST(List, ec_InvalidArgument);

    if (! ttc_mutex_is_locked(&(List->Lock)) ) { // not locked by task: add items to list
        if (List->Pending_Push_Amount)           // still data pending from last isr-push
            _ttc_list_complete_all_pending_isr(List);
        _ttc_list_push_back_multiple(List, FirstItem, LastItem, Amount);
            ttc_semaphore_give(&(List->AmountItems), Amount);
    }
    else {                                       // already locked by task: add to list of pending items
#ifdef TTC_REGRESSION
        LIST_STATS_INCREASE(List, AmountPending);
#endif
        if (List->Pending_Push_Last) { // append at end of list
            List->Pending_Push_Last->Next = FirstItem;
        }
        else {                         // append as first item
            List->Pending_Push_First = FirstItem;
        }
        List->Pending_Push_Last   = LastItem;
        List->Pending_Push_Amount += Amount;

        return tle_ListInUseByTask;
    }

    return tle_OK;
}
     SmallBase_t ttc_list_pop_front_multiple(ttc_list_t* List, SmallBase_t Amount, ttc_list_item_t** FirstItem, ttc_list_item_t** LastItem) {
    Assert_LIST(List,      ec_InvalidArgument);
    Assert_LIST(FirstItem, ec_InvalidArgument);
    Assert_LIST(LastItem,  ec_InvalidArgument);

    while (ttc_list_size(List) == 0) { // list is empty: wait for data
        ttc_semaphore_take(&(List->AmountItems), Amount, -1, ttc_list_pop_front_multiple);
    }

    ttc_mutex_lock(&(List->Lock), -1, ttc_list_pop_front_multiple);
    SmallBase_t AmountRemoved = _ttc_list_pop_front_multiple(List, Amount, FirstItem, LastItem);
    ttc_mutex_unlock(&(List->Lock));

    return AmountRemoved;
}
ttc_list_error_e ttc_list_pop_front_multiple_isr(ttc_list_t* List, SmallBase_t Amount, ttc_list_item_t** FirstItem, ttc_list_item_t** LastItem) {
    Assert_LIST(List,      ec_InvalidArgument);
    Assert_LIST(FirstItem, ec_InvalidArgument);
    Assert_LIST(LastItem,  ec_InvalidArgument);

    if ( ttc_mutex_is_locked(&(List->Lock)) )
        return tle_ListInUseByTask;          // list is locked by a task

    if (List->Pending_Push_Amount)           // still data pending from last isr-push
        _ttc_list_complete_all_pending_isr(List);

    _ttc_list_pop_front_multiple(List, Amount, FirstItem, LastItem);

    return tle_OK;
}
ttc_list_item_t* ttc_list_pop_front_single(ttc_list_t* List) {
    Assert_LIST(List, ec_InvalidArgument);
    Assert_LIST_WRITABLE(List);

    //X while (ttc_list_size(List) == 0) { // list is empty: wait for data
    ttc_semaphore_error_e Error = ttc_semaphore_take(&(List->AmountItems), 1, -1, ttc_list_pop_front_single);
    Assert_LIST(Error == tsme_OK, ec_UNKNOWN);
    //X }
    ttc_list_item_t* Item;
    ttc_mutex_lock(&(List->Lock), -1, ttc_list_pop_front_single);
    _ttc_list_pop_front_single(List, &Item);
    ttc_mutex_unlock(&(List->Lock));
#ifdef TTC_REGRESSION
    Assert_LIST_WRITABLE(List->First);
    Assert_LIST_WRITABLE(Item);
#endif

    return Item;
}
ttc_list_item_t* ttc_list_pop_front_single_isr(ttc_list_t* List) {
    Assert_LIST(List, ec_InvalidArgument);

    // we may use static variables as interrupts are never interrupted!
    static ttc_list_item_t* Item = NULL;

    if ( ttc_mutex_is_locked(&(List->Lock)) )
        return (ttc_list_item_t*) 1;         // list is locked by a task

    if (List->Pending_Push_Amount)           // still data pending from last isr-push
        _ttc_list_complete_all_pending_isr(List);

    if (ttc_semaphore_take(&(List->AmountItems), 1, -1, ttc_list_pop_front_single) != tsme_OK)
        return NULL;

    _ttc_list_pop_front_single(List, &Item);

    return Item;
}

//}functions
//{ private functions *****************************************

       void _ttc_list_complete_all_pending_isr(ttc_list_t* List) {

    _ttc_list_push_back_multiple(List, List->Pending_Push_First, List->Pending_Push_Last, List->Pending_Push_Amount);
    ttc_semaphore_give(&(List->AmountItems), List->Pending_Push_Amount);
    List->Pending_Push_First  = NULL;
    List->Pending_Push_Last   = NULL;
    List->Pending_Push_Amount = 0;
}
       void _ttc_list_push_back_single(ttc_list_t* List, ttc_list_item_t* Item) {
#ifdef TTC_REGRESSION
    Assert_LIST(List, ec_InvalidArgument);
    Assert_LIST(Item, ec_InvalidArgument);
    Assert_LIST(List->LockedBy == tlo_None, ec_InvalidImplementation); // outer lock has been broken
    List->LockedBy = tlo_PushBackSingle;
    Assert_LIST_WRITABLE(List);
    Assert_LIST_WRITABLE(List->First);
    Assert_LIST_WRITABLE(Item);
    List->Stats.Operation = tlo_PushBackSingle;
    ttc_list_add_history(List);
    Assert_LIST(List->Size == ttc_list_count_items(List->First), ec_InvalidImplementation); // bug in size counter implementation
    LIST_STATS_INCREASE(List, AmountPushBack);
#endif

    // push of multiple items would not allow to terminate within constant time!
    Item->Next = TTC_LIST_TERMINATOR;

    if (List->Last) { // list is not empty: append item at end of list
        List->Last->Next = Item;
        List->Last       = Item;
        List->Size++;
    }
    else {            // list is empty: set list of new items as new list
        List->First = Item;
        List->Last  = Item;
        List->Size = 1;
    }

#ifdef TTC_REGRESSION
    Assert_LIST_WRITABLE(Item->Next);
    Assert_LIST_WRITABLE(List->Last->Next);
    List->Stats.Operation = tlo_End;
    ttc_list_add_history(List);
    Assert_LIST_WRITABLE(List->First);
    Assert_LIST_WRITABLE(List->Last);
    Assert_LIST(List->Size == ttc_list_count_items(List->First), ec_InvalidImplementation); // bug in size counter implementation
    List->LockedBy = tlo_None;
#endif
}
       void _ttc_list_push_back_multiple(ttc_list_t* List, ttc_list_item_t* FirstItem, ttc_list_item_t* LastItem, SmallBase_t Amount) {
#ifdef TTC_REGRESSION
    Assert_LIST_WRITABLE(FirstItem);
    List->Stats.Operation = tlo_PushBackSingle;
    LIST_STATS_INCREASE(List, AmountPushBack);
    ttc_list_add_history(List);
#endif

    if (List->Last) { // list is not empty: append item at end of list
        List->Last->Next = FirstItem;
        List->Last       = LastItem;
        List->Size += Amount;
    }
    else {            // list is empty: set list of new items as new list
        List->First = FirstItem;
        List->Last  = LastItem;
        List->Size = Amount;
    }

#ifdef TTC_REGRESSION
    List->Stats.Operation = tlo_End;
    ttc_list_add_history(List);
#endif
}
       void _ttc_list_pop_front_single(ttc_list_t* List, ttc_list_item_t** Item) {
#ifdef TTC_REGRESSION
    Assert_LIST(List, ec_InvalidArgument);
    Assert_LIST(Item, ec_InvalidArgument);
    List->Stats.Operation = tlo_PopFrontSingle;
    ttc_list_add_history(List);
    LIST_STATS_INCREASE(List, AmountPopFront);
    ttc_list_item_t* FirstItem = List->First;
    Assert_LIST(List->Size == ttc_list_count_items(FirstItem), ec_InvalidImplementation); // bug in size counter implementation
    Assert_LIST(List->LockedBy == tlo_None, ec_InvalidImplementation); // outer lock has been broken
    List->LockedBy = tlo_PopFrontSingle;
#endif

    *Item = List->First;
#ifdef TTC_REGRESSION
    Assert_LIST_WRITABLE(*Item);
#endif
    if (*Item) {     // list stores items: remove first one
        List->Size--;
#ifdef TTC_REGRESSION
        Assert_LIST_WRITABLE(List->First->Next);
#endif
        if (List->First->Next == TTC_LIST_TERMINATOR) { // list is empty now
            List->First = NULL;
            List->Last  = NULL;
#ifdef TTC_REGRESSION
            Assert_LIST(List->Size == 0, ec_InvalidImplementation); // bug in size counter implementation
#endif
        }
        else
            List->First = List->First->Next;
    }

#ifdef TTC_REGRESSION
    List->Stats.Operation = tlo_End;
    ttc_list_add_history(List);
    Assert_LIST_WRITABLE(List->First);
    Assert_LIST(List->Size == ttc_list_count_items(List->First), ec_InvalidImplementation); // bug in size counter implementation
    List->LockedBy = tlo_None;
#endif
}
SmallBase_t _ttc_list_pop_front_multiple(ttc_list_t* List, SmallBase_t Amount, ttc_list_item_t** FirstItem, ttc_list_item_t** LastItem) {
#ifdef TTC_REGRESSION
    Assert_LIST(List,      ec_InvalidArgument);
    Assert_LIST(FirstItem, ec_InvalidArgument);
    Assert_LIST(LastItem,  ec_InvalidArgument);
    LIST_STATS_INCREASE(List, AmountPopFront);
#endif

    if (List->First == NULL) { // list is empty
        *FirstItem = NULL;
        *LastItem  = NULL;
        return 0;
    }
    if (Amount == -1) {        // request for all items in list
        Amount     = List->Size;
        *FirstItem = List->First;
        *LastItem  = List->Last;
        List->Size  = 0;
        List->First = NULL;
        List->Last  = NULL;
    }
    else {                     // request for limited amount of items
        if (Amount < List->Size) { // list has less items than requested
            Amount = List->Size;
            List->Size = 0;
        }
        else {                     // list has as much items as requested
            List->Size -= Amount;
        }
        ttc_list_item_t* CurrentItem = List->First;
        *FirstItem = CurrentItem;

        SmallBase_t Count = Amount;
        while (Count-- > 0) { // collect requested amount of items (find last one)
            Assert_LIST_WRITABLE(CurrentItem)

            Assert_LIST(CurrentItem, ec_InvalidImplementation); // somebody has corrupted our list!
            CurrentItem = CurrentItem->Next;
        }
        *LastItem   = CurrentItem;
        List->First = CurrentItem;
        if (CurrentItem == TTC_LIST_TERMINATOR) // list is empty now
            List->Last = NULL;
    }

    return Amount;
}

//} private functions

