#ifndef ttc_list_H
#define ttc_list_H

/** ttc_list.h ***********************************************{
 *
 * Written by Gregor Rebel 2013
 *
 * Architecture independent support for single linked lists.
 * 
 * Why to use lists instead of queues?
 * - lists have no size limit (every item brings its own next-pointer)
 * - lists require only as much memory as required (no oversized pre-allocation like queues)
 * - lists have constant access times too
 *
 * When not to prefer lists over queues?
 * - if you want to store big amount of very small blocks like single bytes
 *   (every item requires a next-pointer)
}*/

//{ includes

#include "ttc_basic.h"
#include "ttc_mutex.h"
#include "ttc_list_types.h"
#include "ttc_memory.h"

//}includes
//{ Default settings (change via makefile) ******************************

#ifndef TTC_Assert_LIST
  #define TTC_Assert_LIST 1
#endif

//}Default settings
//{ Types & Structures ***************************************************

#if TTC_Assert_LIST == 1 // use Assert()s in List code (somewhat slower but alot easier to debug)
#  define Assert_LIST(Condition, ErrorCode) Assert(Condition, ErrorCode)
#  define Assert_LIST_WRITABLE(ADDRESS) if (ADDRESS && (((ttc_list_item_t*) ADDRESS)!=TTC_LIST_TERMINATOR)) { Assert(ttc_memory_is_writable(ADDRESS), ec_InvalidImplementation); }
#else  // us no Assert()s in USART code (somewhat smaller + faster, but crashes are hard to debug)
#  define Assert_LIST(Condition, ErrorCode)
#  define Assert_LIST_WRITABLE(ADDRESS)
#endif

//}
//{ Function prototypes **************************************************

/** counts amount of items linked together starting at given item
 *
 * Note: This functions terminates within O(n) with n = amount of items
 *
 * @param FirstItem   list item to start counting at
 * @return            amount of items linked together until first NULL-item
 */
Base_t ttc_list_count_items(ttc_list_item_t* FirstItem);

/** allocates + initializes a new list
 *
 * @param  Name  used for debugging purposes only
 * @return =! NULL: list has been created successfully (use handle to operate on list)
 */
ttc_list_t* ttc_list_create(const char* Name);

/** initializes a new list in given storage place
 *
 * @param List   pointer to memory space to be used to manage list
 * @param  Name  used for debugging purposes only
 * @return =! NULL: list has been created successfully (use handle to operate on list)
 */
void ttc_list_init(ttc_list_t* List, const char* Name);

/** returns amount of items stored in List
 *
 * Note: This function completes in constant time O(1)
 *
 * @param List     pointer to already created list
 * @return         Amount of items currently stored in List
 */
// ttc_ListSize_t always uses native data width of current architecture
#define ttc_list_size(LIST)  LIST->Size
//Base_t ttc_list_size(ttc_list_t* List);

/** completes all pending operations on given list
 *
 * Note: This function may only be called from an interrupt service routine only!
 * Note: This function completes in constant time O(1)
 *
 * @param List     pointer to already created list
 * @return         == tle_OK: all pending operations have been completed
 *                 == tle_ListInUseByTask: list is locked by a task (no operations completed)
 */
ttc_list_error_e ttc_list_complete_all_pending_isr(ttc_list_t* List);

/** adds given single Item at end of given list
 * @param List     pointer to already created list
 * @param Item     pointer to item to be added to list
 */
void ttc_list_push_back_single(ttc_list_t* List, ttc_list_item_t* Item);

/** adds given linked list of items at end of given list
 * @param List       pointer to already created list
 * @param FirstItem  pointer to first item in list of items to be added to list
 * @param LastItem   pointer to last item in list of items to be added to list
 * @param Amount     Amount of items to be added to list
 */
void ttc_list_push_back_multiple(ttc_list_t* List, ttc_list_item_t* FirstItem, ttc_list_item_t* LastItem, SmallBase_t Amount);

/** adds given single Item at end of given list (does never block)
 *
 * Note: Must be called from interrupt service routine only!
 * Note: function terminates in O(1)
 *
 * @param List     pointer to already created list
 * @param Item     pointer to item to be added to list
 * @return         == tle_OK: Item has been successfully added to List
 *                 == tle_ListInUseByTask: items have been added to pending list and will be later on later calls
 *                                         (you may call ttc_list_push_back_isr(List, NULL) to retry push
 */
ttc_list_error_e ttc_list_push_back_single_isr(ttc_list_t* List, ttc_list_item_t* Item);

/** adds given linked list of Items at end of given list (does never block)
 *
 * Note: Must be called from interrupt service routine only!
 * Note: function terminates in O(1)
 *
 * @param List       pointer to already created list
 * @param FirstItem  pointer to first item in list of items to be added to list
 * @param LastItem   pointer to last item in list of items to be added to list
 * @param Amount     Amount of items to be added to list
 * @return         == tle_OK: Item has been successfully added to List
 *                 == tle_ListInUseByTask: items have been added to pending list and will be later on later calls
 *                                         (you may call ttc_list_push_back_isr(List, NULL) to retry push
 */
ttc_list_error_e ttc_list_push_back_multiple_isr(ttc_list_t* List, ttc_list_item_t* FirstItem, ttc_list_item_t* LastItem, SmallBase_t Amount);

/** removes first Item from given list and delivers its reference
 * @param List     pointer to already created list
 * @return         != NULL: pointer to item removed from given list; == NULL: list is empty
 */
ttc_list_item_t* ttc_list_pop_front_single(ttc_list_t* List);

/** removes linked list of first items from front of given list and delivers reference to first and last one
 *
 * Note: This function completes in constant time O(Amount)
 *
 * @param List      pointer to already created list
 * @param Amount    amount of items to remove from list (==-1: means all items in list)
 * @param FirstItem loaded with pointer to first item of items removed from given list or NULL if list is empty
 * @param LastItem  loaded with pointer to last item of items removed from given list or NULL if list is empty
 * @return          actual amount of items being removed from list
 */
SmallBase_t ttc_list_pop_front_multiple(ttc_list_t* List, SmallBase_t Amount, ttc_list_item_t** FirstItem, ttc_list_item_t** LastItem);

/** removes linked list of first items from front of given list and delivers reference to first and last one (does never block)
 *
 * Note: Must be called from interrupt service routine only!
 * Note: This function completes in constant time O(Amount)
 *
 * @param List      pointer to already created list
 * @param Amount    amount of items to remove from list (==-1: means all items in list)
 * @param FirstItem loaded with pointer to first item of items removed from given list or NULL if list is empty
 * @param LastItem  loaded with pointer to last item of items removed from given list or NULL if list is empty
 * @return         == tle_OK: Items have been successfully removed from list
 *                 == tle_ListInUseByTask: list is currently in use by task; try again later
 */
ttc_list_error_e ttc_list_pop_front_multiple_isr(ttc_list_t* List, SmallBase_t Amount, ttc_list_item_t** FirstItem, ttc_list_item_t** LastItem);

/** removes first Items from given list and delivers reference to first item (does never block)
 *
 * Note: Must be called from interrupt service routine only!
 *
 * @param List     pointer to already created list
 * @return         != NULL: pointer to item removed from given list;
 *                 == NULL: list is empty
 *                 == 1:    list is locked by task
 */
ttc_list_item_t* ttc_list_pop_front_single_isr(ttc_list_t* List);

//}Function prototypes

#endif
