#ifndef ttc_mutex_H
#define ttc_mutex_H

/** ttc_mutex.h ***********************************************{
 * 
 * Written by Gregor Rebel 2010-2013
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported architecture/ schedulers: CortexM3 (cm3), FreeRTOS
 * 
 * Why using ttc_mutex.c instead of cm3_mutex.c or FreeRTOS-functions directly?
 * (1) it is more portable
 *     You may change to another multitasking scheduler once it is supported by The ToolChain
 * (2) extra safety checks
 *     Deadlocks are a common issue in multitasking applications and sometimes hard to debug.
 *     + ttc_mutex allows to turn all endless locks into a timed one that will assert if expired.
 *       -> ttc_mutex_types.h/TTC_MUTEX_TIMEOUT_USECS
 *     + ttc_mutex provides smart mutexes that can record history of operations and
 *       run as fast as normal mutexs if disabled.
 *       -> ttc_mutex_types.h/TTC_SMART_MUTEXES
 *
 * Smart mutexes store the owner of a lock.
 * Whenever a mutex stays locked because of a programming error, then the last locking
 * function can be identified by taking a look into the smart lock structure.
}*/

//{ includes

#include "ttc_basic.h"

#ifdef EXTENSION_500_ttc_mutex_cortexm3
#  include "cm3_mutex.h"
#endif

#ifdef EXTENSION_500_ttc_mutex_freertos
#  include "freertos_mutex.h"
#endif

#include "ttc_mutex_types.h"
#ifdef EXTENSION_500_ttc_task
#  include "ttc_task.h"
#endif

//}includes
//{ Types & Structures ***************************************************

/** Types to be defined by low-level driver ******************************
 *
 * ttc_mutex_t  datatype storing a mutex
 * E.g.: typedef u8_t ttc_mutex_t
 *
 *
 */


//}
#include "ttc_memory.h"

//{ Function prototypes **************************************************
//
// For each function, a DRIVER_() pendant must be defined by low-level driver to provide an implementation.


/** Checks if given mutex is in locked state
 *
 * Note: Calling this function makes only sense from interrupt service routine
 *
 * @param Mutex  handle of already created mutex
 * @return         == tme_OK: Lock has been released successfully
 *                 == tme_WasLockedByTask: call task ttc_task_yield_isr() before exiting your ISR!
 */
BOOL ttc_mutex_is_locked(ttc_mutex_smart_t* Mutex);  // just a prototype for easier understanding
#if TTC_SMART_MUTEXES == 1
#  define ttc_mutex_is_locked(Mutex) _ttc_mutex_smart_is_locked(Mutex)
#else
#  define ttc_mutex_is_locked(Mutex) _ttc_mutex_is_locked(Mutex)
#endif


/** creates a new non-recursive mutex lock
 * @return                 =! NULL: mutex has been created successfully (use handle to operate on mutex)
 */
ttc_mutex_smart_t* ttc_mutex_create();  // just a prototype for easier understanding
#if TTC_SMART_MUTEXES == 1
#define ttc_mutex_create()   _ttc_mutex_smart_create()
#else
#  define ttc_mutex_create() _ttc_mutex_create()
#endif

/** initializes an existing non-recursive mutex lock to the unlocked state
 * @Mutex                 storage to be initialized as mutex
 */
void ttc_mutex_init(ttc_mutex_smart_t* Lock);  // just a prototype for easier understanding
#if TTC_SMART_MUTEXES == 1
#  define ttc_mutex_init(Mutex) _ttc_mutex_smart_init(Mutex)
#else
#  define ttc_mutex_init(Mutex) _ttc_mutex_init(Mutex)
#endif

/** Tries to obtain lock on mutex within timeout
 * @param Mutex  handle of already created mutex
 * @param TimeOut  !=-1: amount of system ticks to wait for lock
 *                 ==-1: wait forever
 * @param Owner    reference to function which wants to obtain the lock
 * @return         == tme_OK: Lock has been successfully obtained within TimeOut
 */
ttc_mutex_error_e ttc_mutex_lock(ttc_mutex_smart_t* Mutex, Base_t TimeOut, void (*Owner)() );  // just a prototype for easier understanding
#if TTC_SMART_MUTEXES == 1
#  define ttc_mutex_lock(Mutex, TimeOut, Owner) _ttc_mutex_smart_lock(Mutex, TimeOut, (void (*)()) Owner)
#else
#  define ttc_mutex_lock(Mutex, TimeOut, Owner) _ttc_mutex_lock(Mutex, TimeOut)
#endif

/** Tries to obtain lock on mutex; this function will never block
 *
 * Note: This function must only be used with disabled interrupts (e.g. from interrupt service routine)!
 *
 * @param Mutex    handle of already created mutex
 * @param Owner    reference to function which wants to obtain the lock
 * @return         == tme_OK: Lock has been successfully obtained within TimeOut
 *                 == tme_WasLockedByTask: - a task of higher priority has been interrupted => call ttc_task_yield() at end of your interrupt service routine!
 *                 >= tme_Error: Error condition
 */
ttc_mutex_error_e ttc_mutex_lock_isr(ttc_mutex_smart_t* Mutex, void (*Owner)() );  // just a prototype for easier understanding
#ifndef DRIVER_ttc_mutex_lock_isr
#error Missing low-level implementation! // check description above for function to be implemented
#endif
#if TTC_SMART_MUTEXES == 1
#  define ttc_mutex_lock_isr(Mutex, Owner) _ttc_mutex_smart_lock_isr(Mutex, (void (*)()) Owner)
#else
#  define ttc_mutex_lock_isr(Mutex, Owner) _ttc_mutex_lock_isr(Mutex)
#endif

/** Releases an already obtained lock from mutex
 * @param Mutex  handle of already created smart mutex
 * @return         == tme_OK: Lock has been released successfully
 */
ttc_mutex_error_e ttc_mutex_unlock(ttc_mutex_smart_t* Mutex);  // just a prototype for easier understanding
#if TTC_SMART_MUTEXES == 1
#define ttc_mutex_unlock(Mutex) _ttc_mutex_smart_unlock(Mutex)
#else
#define ttc_mutex_unlock(Mutex) _ttc_mutex_unlock(Mutex)
#endif


/** Releases an already obtained lock from mutex (does never block)
 * @param Mutex  handle of already created mutex
 * @return         == tme_OK: Lock has been released successfully
 *                 == tme_WasLockedByTask: call task ttc_task_yield_isr() before exiting your ISR!
 */
ttc_mutex_error_e ttc_mutex_unlock_isr(ttc_mutex_smart_t* Mutex);  // just a prototype for easier understanding
#if TTC_SMART_MUTEXES == 1
#  define ttc_mutex_unlock_isr(Mutex) _ttc_mutex_smart_unlock_isr(Mutex)
#else
#  define ttc_mutex_unlock_isr(Mutex) _ttc_mutex_unlock_isr(Mutex)
#endif

//}Function prototypes
//{ low-level prototypes (call via macros above!)

ttc_mutex_t* _ttc_mutex_create();
ttc_mutex_smart_t* _ttc_mutex_smart_create();
#ifndef DRIVER_ttc_mutex_create
#error Missing low-level implementation! // check description above for function to be implemented
#endif

void _ttc_mutex_init(ttc_mutex_t* Lock);
void _ttc_mutex_smart_init(ttc_mutex_smart_t* Mutex);
#ifndef DRIVER_ttc_mutex_init
#error Missing low-level implementation! // check description above for function to be implemented
#endif

ttc_mutex_error_e _ttc_mutex_lock(ttc_mutex_t* Mutex, Base_t TimeOut);
ttc_mutex_error_e _ttc_mutex_smart_lock(ttc_mutex_smart_t* SmartMutex, Base_t TimeOut, void (*Owner)() );
#ifndef DRIVER_ttc_mutex_lock
#error Missing low-level implementation! // check description above for function to be implemented
#endif

ttc_mutex_error_e _ttc_mutex_lock_isr(ttc_mutex_t* Mutex);
ttc_mutex_error_e _ttc_mutex_smart_lock_isr(ttc_mutex_smart_t* SmartMutex, void (*Owner)() );
#ifndef DRIVER_ttc_mutex_lock_isr
#error Missing low-level implementation! // check description above for function to be implemented
#endif

BOOL _ttc_mutex_is_locked(ttc_mutex_t* Mutex);
BOOL _ttc_mutex_smart_is_locked(ttc_mutex_smart_t* SmartMutex);
#ifndef DRIVER_ttc_mutex_is_locked
#error Missing low-level implementation! // check description above for function to be implemented
#endif

ttc_mutex_error_e _ttc_mutex_unlock(ttc_mutex_t* Mutex);
ttc_mutex_error_e _ttc_mutex_smart_unlock(ttc_mutex_smart_t* SmartMutex);
#ifndef DRIVER_ttc_mutex_unlock
#error Missing low-level implementation! // check description above for function to be implemented
#endif

ttc_mutex_error_e _ttc_mutex_unlock_isr(ttc_mutex_t* Mutex);
ttc_mutex_error_e _ttc_mutex_smart_unlock_isr(ttc_mutex_smart_t* SmartMutex);
#ifndef DRIVER_ttc_mutex_unlock_isr
#error Missing low-level implementation! // check description above for function to be implemented
#endif

//} low-level prototypes


#endif
