#ifndef ttc_mutex_types_H
#define ttc_mutex_types_H

/*{ ttc_mutex_types.h ***********************************************
 * 
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported architectures/ schedulers: CortexM3, FreeRTOS
 * 
}*/

//{ includes

#include "ttc_basic.h"
#include "ttc_task_types.h"

// include low-level driver types
//
// Required defines/ typedefs:
// ttc_mutex_t   must be a typedef to datatype to store data of a single mutex on current architecture

#ifdef EXTENSION_500_ttc_mutex_cortexm3
#include "cm3_mutex_types.h"
#endif


#ifdef EXTENSION_500_ttc_mutex_freertos
#  include "freertos_mutex_types.h"
#endif

#ifndef ttc_mutex_t
//? #error Missing low-level definition for ttc_mutex_t
#endif

//? typedef void* ttc_QueueHandle;

typedef enum ttc_mutex_error_E {      // return type of most functions in ttc_mutex
    tme_OK,                           // no error
    tme_WasLockedByTask,              // an _isr()-call occured while Mutex was locked by a task

    tme_Error,
    // Error Conditions below
    tme_TimeOut,                      // an operation has timed out
    tme_NotImplemented,               // required function not implemented for current architecture
    tme_MutexAlreadyLocked,                  // tried to lock a mutex that is already locked
    tme_UNKNOWN
} ttc_mutex_error_e;

//}includes
//{ Default settings (change via makefile) ******************************


#ifndef TTC_MUTEX_TIMEOUT_USECS
  #define TTC_MUTEX_TIMEOUT_USECS 0 //1000000 // >0: activate timeout-assert for endless mutexes (TimeOut==-1) in usecs
  // #define TTC_MUTEX_TIMEOUT_USECS 0    // no forced timeout
#endif

// enable use of Assert() for mutexes
#ifndef TTC_ASSERT_MUTEXES
  #define TTC_ASSERT_MUTEXES 1
#endif

//}Default settings
//{ Types & Structures ***************************************************

#if TTC_ASSERT_MUTEXES == 1 // use Assert()s in Queue code (somewhat slower but alot easier to debug)
  #define Assert_MUTEXES(Condition, ErrorCode) Assert(Condition, ErrorCode)
#else  // us no Assert()s in USART code (somewhat smaller + faster, but crashes are hard to debug)
  #define Assert_MUTEXES(Condition, ErrorCode)
#endif

// enable use of smart mutexes
#ifndef TTC_SMART_MUTEXES
  #define TTC_SMART_MUTEXES 1
#endif

// define amount
#if TTC_SMART_MUTEXES == 1

  // set amount of entries to be stored in history of smart mutex usage (debugging only)
  #ifndef TTC_SMART_MUTEX_HISTORY_SIZE
    #define TTC_SMART_MUTEX_HISTORY_SIZE 0
  #endif

#endif

// debuggable mutex that can store its owner
// This comes handy if you wanna know, who has locked a mutex before.
#if TTC_SMART_MUTEXES == 1
  typedef struct ttc_mutex_smart_s {
    ttc_mutex_t Mutex;

    // function pointer given to lock() function
    void (*Owner)();

    // task info of owner
    ttc_task_info_t* OwnerTask;

  } ttc_mutex_smart_t;
#else
  typedef ttc_mutex_t ttc_mutex_smart_t;
#endif

//}

/** called while waiting for mutex lock. Place your breakpoint inside to debug hanging locks
  */
void _ttc_mutex_wait();



#endif
