/*{ ttc_queue.c ***********************************************
 *
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported schedulers: FreeRTOS
}*/
#include "ttc_queue.h"

// private functions ********************************************

// variables used by different interrupt service routines
// declaring their local variables as global saves stack memory
static u8_t   tqpp_AmountPushed;
static u8_t   tqpp_AmountRemaining;
static ttc_queue_error_e ttcq_QueueError;

ttc_queue_error_e _ttc_queue_pointer_push_back(ttc_queue_pointers_t* Queue, void* Data) {

    Base_t NewEnd = Queue->End + 1;
    if (NewEnd >= Queue->Size)
        NewEnd = 0;                 // wrapping around at end of array

    if (Queue->Start == NewEnd) {
        return tqe_QueuePushFailed; // queue is full
    }

    Queue->Data[Queue->End] = Data; // storing at old end index
    Queue->End = NewEnd;

    return tqe_OK;
}
ttc_queue_error_e _ttc_queue_pointer_pull_front(ttc_queue_pointers_t* Queue, void** Data) {

    u16_t Start = Queue->Start;
    Assert_QUEUES(Start < Queue->Size,  ec_InvalidImplementation);

    if (Start == Queue->End) {   // queue is empty
        *Data = NULL;
        return tqe_QueueIsEmpty;
    }
    void* Address = Queue->Data[Start++];
    *Data = Address;

    if (Start == Queue->Size)
        Start = 0;              // wrapping around at end of array
    Queue->Start = Start;

    return tqe_OK;
}
             void _ttc_queue_pointer_process_pending(ttc_queue_pointers_t* Queue) {

    // function gets called from interrupt service routine only: no need to waist stack memory
    static void** Data;
    static void** Writer;

    Data = Queue->PendingPush;
    tqpp_AmountPushed = 0;

    while (Queue->AmountPendingPushes) { // process pending pushes
        if (_ttc_queue_pointer_push_back(Queue, *Data++) > tqe_Error) { // cannot process all pending: cleanup
            tqpp_AmountRemaining = Queue->AmountPendingPushes;
            Writer = Queue->PendingPush;
            Data--; // back to last pending Data
            while (tqpp_AmountRemaining--) {
                *Writer++ = *Data++;
            }
            break;
        }
        Queue->AmountPendingPushes--;
        tqpp_AmountPushed++;
    }
    if (tqpp_AmountPushed) {
        ttc_semaphore_give_isr( &(Queue->Amount), tqpp_AmountPushed);
    }
}
ttc_queue_error_e _ttc_queue_byte_try_push_back(ttc_queue_bytes_t* Queue, u8_t Data) {

    Base_t NewEnd = Queue->End + 1;
    if (NewEnd >= Queue->Size)
        NewEnd = 0;

    if (Queue->Start == NewEnd) {
        return tqe_QueuePushFailed; // queue is full
    }

    Queue->Data[Queue->End] = Data; // storing at old end index
    Queue->End = NewEnd;

    return tqe_OK;
}
ttc_queue_error_e _ttc_queue_byte_pull_front(ttc_queue_bytes_t* Queue, u8_t* Data) {
    u16_t Start = Queue->Start;
    Assert_QUEUES(Start < Queue->Size,  ec_UNKNOWN);

    if (Start == Queue->End) {      // queue is empty
        return tqe_QueueIsEmpty;
    }

    u8_t Byte = Queue->Data[Start++];
    *Data = Byte;

    if (Start == Queue->Size)
        Start = 0;                 // wrapping around at end of array
    Queue->Start = Start;

    return tqe_OK;
}
             void _ttc_queue_byte_process_pending(ttc_queue_bytes_t* Queue) {

    // function gets called from interrupt service routine only: no need to waist stack memory
    static u8_t* Data;
    static u8_t* Writer;

    Data = Queue->PendingPush;
    tqpp_AmountPushed = 0;

    while (Queue->AmountPendingPushes) { // process pending pushes
        if (_ttc_queue_byte_try_push_back(Queue, *Data++) > tqe_Error) { // cannot process all pending: cleanup
            tqpp_AmountRemaining = Queue->AmountPendingPushes;
            Writer = Queue->PendingPush;
            Data--; // back to last pending Data
            while (tqpp_AmountRemaining--) {
                *Writer++ = *Data++;
            }
            break;
        }
        Queue->AmountPendingPushes--;
        tqpp_AmountPushed++;
    }
    if (tqpp_AmountPushed) {
        ttc_semaphore_give_isr( &(Queue->Amount), tqpp_AmountPushed);
    }
}

//}functions
//{ functions *************************************************

// generic queue (external implementation based on scheduler)
      ttc_QueueHandle ttc_queue_create(Base_t AmountOfEntries, Base_t SizeOfEntries) {
    ttc_QueueHandle NewQueue = NULL;

#ifdef EXTENSION_300_scheduler_free_rtos
    NewQueue = (ttc_QueueHandle) xQueueCreate(AmountOfEntries, SizeOfEntries);
#endif

    Assert_QUEUES(NewQueue != NULL, ec_Malloc); // could not create queue (no implementation/ not egnough memory)

    return NewQueue;
}
    ttc_queue_error_e ttc_queue_push_back(ttc_QueueHandle Queue, const void* Entry, Base_t TimeOut) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);
    Assert_QUEUES(ttc_task_is_scheduler_running(), ec_InvalidConfiguration); // start scheduler before using this function!

#ifdef EXTENSION_300_scheduler_free_rtos
    if (xQueueSend( (xQueueHandle) Queue, Entry, TimeOut ) == pdTRUE)
        return tqe_OK;
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
    ttc_queue_error_e ttc_queue_push_back_isr(ttc_QueueHandle Queue, const void* Entry) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
    static signed portBASE_TYPE Status;
    Status = xQueueSendToBackFromISR( (xQueueHandle) Queue, Entry, &xHigherPriorityTaskWoken);

    if (xHigherPriorityTaskWoken == pdFALSE) {
        if (Status == pdPASS)
            return tqe_OK;
        else
            return tqe_QueuePushFailed;
    }
    else
        return tqe_HigherPriorityTaskWoken;
#endif

    return tqe_NotImplemented;
}
    ttc_queue_error_e ttc_queue_pop_front(ttc_QueueHandle Queue, void* Entry, Base_t TimeOut) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);
    Assert_QUEUES(ttc_task_is_scheduler_running(), ec_InvalidConfiguration); // start scheduler before using this function!

#ifdef EXTENSION_300_scheduler_free_rtos
    if ( xQueueReceive( (xQueueHandle) Queue, Entry, TimeOut ) )
        return tqe_OK;
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
    ttc_queue_error_e ttc_queue_pop_front_isr(ttc_QueueHandle Queue, void* Entry) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    static signed portBASE_TYPE pxTaskWoken;
    pxTaskWoken = 0;
    if ( xQueueReceiveFromISR( (xQueueHandle) Queue, Entry, &pxTaskWoken) ) {
        if (pxTaskWoken)
            return tqe_HigherPriorityTaskWoken;
        return tqe_OK;
    }
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
               Base_t ttc_queue_amount_waiting(ttc_QueueHandle Queue) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(ttc_task_is_scheduler_running(), ec_InvalidConfiguration); // start scheduler before using this function!

#ifdef EXTENSION_300_scheduler_free_rtos
    return uxQueueMessagesWaiting( (xQueueHandle) Queue );
#endif

    return 0;
}

// simple byte queue (fast + low memory usage + independent from scheduler)
   ttc_queue_bytes_t* ttc_queue_byte_create(u16_t AmountOfEntries) {
    Assert_QUEUES(AmountOfEntries, ec_InvalidArgument); // need at least size 1

    // need to allocate one additional entry for queue-management
    AmountOfEntries++;

    ttc_queue_bytes_t* Queue = ttc_memory_alloc(AmountOfEntries + sizeof(ttc_queue_bytes_t));
    ttc_queue_byte_init(Queue, (u8_t*) (Queue + 1), AmountOfEntries);

    return Queue;
}
                 void ttc_queue_byte_init( ttc_queue_bytes_t* Queue, u8_t* Storage, u16_t AmountOfEntries) {
    Assert_QUEUES(AmountOfEntries, ec_InvalidArgument); // need at least size 1

    ttc_memory_set(Queue, 0, sizeof(*Queue));
    Queue->Size  = AmountOfEntries;
    ttc_mutex_init( &(Queue->Lock) );
    ttc_semaphore_init( &(Queue->Amount) );
    Queue->Data  = Storage;
                 }
               Base_t ttc_queue_byte_get_amount(ttc_queue_bytes_t* Queue) {
    Assert_QUEUES(Queue, ec_InvalidArgument);

    Base_t Amount = ttc_semaphore_available(&(Queue->Amount));

#ifdef TTC_REGRESSION
    if (Queue->End < Queue->Start) {
        Assert(Amount == Queue->End + Queue->Size - Queue->Start, ec_InvalidImplementation);
    }
    else {
        Assert(Amount == Queue->End - Queue->Start, ec_InvalidImplementation);
    }
#endif

    return Amount + Queue->AmountPendingPushes;
}
                 void ttc_queue_byte_push_back(ttc_queue_bytes_t* Queue, u8_t Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);
    ttc_queue_error_e Error = tqe_UNKNOWN;

    do {
        ttc_mutex_lock(&(Queue->Lock), -1, ttc_queue_byte_push_back);
        if (Queue->AmountPendingPushes) { // data from isr is pending to be pushed: complete push
            ttc_task_begin_criticalsection(ttc_queue_byte_pull_front);
            ttc_interrupt_all_disable();
            _ttc_queue_byte_process_pending(Queue);
            ttc_interrupt_all_enable();
            ttc_task_end_criticalsection();
        }
        Error = _ttc_queue_byte_try_push_back(Queue, Data);
        if (Error != tqe_OK) {    // Queue full: append ourself to list of waiting tasks
            /* Note: Write access to Queue->FirstWaitingTask/ LastWaitingTask is allowed only here! */

            ttc_queue_waiting_task_t ThisWaitingTask;
            ThisWaitingTask.Next = NULL;
            ThisWaitingTask.Task = ttc_task_current_handle();
            if (!Queue->FirstTaskWaiting2Push) { // no other task is waiting to push: make us first task
                 Queue->FirstTaskWaiting2Push = &ThisWaitingTask;
                 Queue->LastTaskWaiting2Push  = &ThisWaitingTask;
            }
            else {                               // append ourself on end of list of waiting tasks
#ifdef TTC_REGRESSION
                Assert_QUEUES(ttc_memory_is_writable(Queue->FirstTaskWaiting2Push), ec_InvalidImplementation); // error is list data
                Assert_QUEUES(ttc_memory_is_writable(Queue->LastTaskWaiting2Push),  ec_InvalidImplementation); // error is list data
#endif
                Queue->LastTaskWaiting2Push->Next = &ThisWaitingTask;
                Queue->LastTaskWaiting2Push       = &ThisWaitingTask;
            }
#ifdef TTC_REGRESSION
            Queue->AmountTaskWaitingTotal++;
#endif
            ttc_mutex_unlock( &(Queue->Lock) );
            ttc_task_suspend(NULL); // put ourself to sleep

            // got woken up: remove ourself from list of waiting tasks
            // Note: if Queue->FirstWaitingTask gets NULL, then LastTaskWaiting2Push is invalid!
            Queue->FirstTaskWaiting2Push = ThisWaitingTask.Next;

            Error = tqe_QueuePushFailed;
        }
        else {                    // push successfull
            ttc_mutex_unlock( &(Queue->Lock) );
        }
    } while (Error != tqe_OK);
    ttc_semaphore_give( &(Queue->Amount), 1);
}
    ttc_queue_error_e ttc_queue_byte_push_back_isr(ttc_queue_bytes_t* Queue, u8_t Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);

    if (! ttc_mutex_is_locked( &(Queue->Lock) ) ) {
        if (Queue->AmountPendingPushes)
            _ttc_queue_byte_process_pending(Queue);
        ttcq_QueueError = _ttc_queue_byte_try_push_back(Queue, Data);
        if (ttcq_QueueError == tqe_OK) { // push successfull: increase value
            ttc_semaphore_give( &(Queue->Amount), 1);
        }
    }
    else { // queue is currently locked by a task: store pending byte in extra isr-storage
        if (Queue->AmountPendingPushes < TTC_SQB_ISR_SPACE) {
            Queue->PendingPush[Queue->AmountPendingPushes++] = Data;
            ttcq_QueueError = tqe_IsrOperationPending;
        }
        else // no more space left for pending pushes
            ttcq_QueueError = tqe_QueueIsLocked;
#ifdef TTC_REGRESSION
        Queue->AmountPendingTotal++;
#endif
    }

    return ttcq_QueueError;
}
    ttc_queue_error_e ttc_queue_byte_pull_front(ttc_queue_bytes_t* Queue, u8_t* Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(Data,  ec_InvalidArgument);

    ttc_semaphore_take( &(Queue->Amount), 1, -1, ttc_queue_byte_pull_front);
    ttc_mutex_lock( &(Queue->Lock), -1, ttc_queue_byte_pull_front);
    ttc_queue_error_e Error = _ttc_queue_byte_pull_front(Queue, Data);

    if (Queue->AmountPendingPushes)                              { // data from isr is pending to be pushed: complete push
        ttc_task_begin_criticalsection(ttc_queue_byte_pull_front);
        ttc_interrupt_all_disable();
        _ttc_queue_byte_process_pending(Queue);
        ttc_mutex_unlock( &(Queue->Lock) );
        ttc_interrupt_all_enable();
        ttc_task_end_criticalsection();
    }
    else if ( (Error == tqe_OK) && Queue->FirstTaskWaiting2Push) { // at least one task is waiting to push: awake it
        if (ttc_semaphore_available(&(Queue->Amount)) < Queue->Size) { // optimization: do not resume pushing task at every byte
            ttc_task_begin_criticalsection(ttc_queue_byte_pull_front);
            ttc_mutex_unlock( &(Queue->Lock) );
            ttc_task_resume(Queue->FirstTaskWaiting2Push->Task);
            ttc_task_end_criticalsection();
        }
        else {
            ttc_mutex_unlock( &(Queue->Lock) );
        }
    }
    else {                                                         // no task waiting to push: release queue
        ttc_mutex_unlock( &(Queue->Lock) );
    }
    if (Error > tqe_Error) { // could not read data: give back token
        ttc_semaphore_give( &(Queue->Amount), 1);
    }

    return Error;
}
    ttc_queue_error_e ttc_queue_byte_pull_front_isr(ttc_queue_bytes_t* Queue, u8_t* Data) {
    // variables may be static because we don't expect concurrency inside ISR
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(Data,  ec_InvalidArgument);

    if (ttc_semaphore_available(&(Queue->Amount))) { // queue stores data: try to pull one
        if (! ttc_mutex_is_locked( &(Queue->Lock) ) ) {
            if (Queue->AmountPendingPushes)
                _ttc_queue_byte_process_pending(Queue);
            ttcq_QueueError = _ttc_queue_byte_pull_front(Queue, Data);
            if (ttcq_QueueError == tqe_OK) {
                ttc_semaphore_take_isr(&(Queue->Amount), 1, ttc_queue_byte_pull_front_isr);
            }
        }
        else { // queue is currently locked by a task
            ttcq_QueueError = tqe_QueueIsLocked;
        }
        return ttcq_QueueError;
    }
    return tqe_QueueIsEmpty;
  }
    ttc_queue_error_e ttc_queue_byte_try_pull_front(ttc_queue_bytes_t* Queue, u8_t* Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(Data,  ec_InvalidArgument);

    ttc_queue_error_e Error = ttc_semaphore_try_take( &(Queue->Amount), 1, ttc_queue_byte_try_pull_front);
    if (Error == tqe_OK) {
        ttc_mutex_lock( &(Queue->Lock) , -1, ttc_queue_byte_try_pull_front);
        Error = _ttc_queue_byte_pull_front(Queue, Data);
        ttc_mutex_unlock( &(Queue->Lock) );
        if (Error > tqe_Error) { // could not read data: give back token
            ttc_semaphore_give( &(Queue->Amount), 1);
        }
    }

    return Error;
}

// simple pointer queue (fast + low memory usage + independent from scheduler)
ttc_queue_pointers_t* ttc_queue_pointer_create(u16_t AmountOfEntries) {
    Assert_QUEUES(AmountOfEntries, ec_InvalidArgument); // need at least size 1

    // need to allocate one additional entry for queue-management
    AmountOfEntries++;

    ttc_queue_pointers_t* Queue = ttc_memory_alloc(sizeof(void*) * AmountOfEntries + sizeof(ttc_queue_pointers_t));
    ttc_queue_pointer_init(Queue, (void**) (Queue + 1), AmountOfEntries);

    return Queue;
}
                 void ttc_queue_pointer_init(ttc_queue_pointers_t* Queue, void** Storage, u16_t AmountOfEntries) {
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(AmountOfEntries, ec_InvalidArgument); // need at least size 1

    ttc_memory_set(Queue, 0, sizeof(*Queue));

    Queue->Size  = AmountOfEntries;
    ttc_mutex_init( &(Queue->Lock) );
    ttc_semaphore_init( &(Queue->Amount) );
    Queue->Data  = Storage;
}
               Base_t ttc_queue_pointer_get_amount(ttc_queue_pointers_t* Queue) {
    Assert_QUEUES(Queue, ec_InvalidArgument);

    Base_t Amount = ttc_semaphore_available(&(Queue->Amount));

#ifdef TTC_REGRESSION
    if (Queue->End < Queue->Start) {
        Assert(Amount == Queue->End + Queue->Size - Queue->Start, ec_InvalidImplementation);
    }
    else {
        Assert(Amount == Queue->End - Queue->Start, ec_InvalidImplementation);
    }
#endif

    return Amount + Queue->AmountPendingPushes;
}
    ttc_queue_error_e ttc_queue_pointer_push_back(ttc_queue_pointers_t* Queue, void* Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);
#ifdef TTC_REGRESSION
    Assert_QUEUES(ttc_memory_is_writable(Data), ec_InvalidImplementation);
#endif

    ttc_mutex_lock(&(Queue->Lock), -1, ttc_queue_pointer_push_back);
    if (Queue->AmountPendingPushes) { // data from isr is pending to be pushed: complete push
        ttc_task_begin_criticalsection(ttc_queue_byte_pull_front);
        ttc_interrupt_all_disable();
        _ttc_queue_pointer_process_pending(Queue);
        ttc_interrupt_all_enable();
        ttc_task_end_criticalsection();
    }
    ttc_queue_error_e Error = _ttc_queue_pointer_push_back(Queue, Data);
    ttc_mutex_unlock( &(Queue->Lock) );
    if (Error == tqe_OK) { // push successfull: increase value
        ttc_semaphore_give( &(Queue->Amount), 1);
    }

    return Error;
}
    ttc_queue_error_e ttc_queue_pointer_push_back_isr(ttc_queue_pointers_t* Queue, void* Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);

    if (! ttc_mutex_is_locked( &(Queue->Lock) ) ) {
        if (Queue->AmountPendingPushes)
            _ttc_queue_pointer_process_pending(Queue);
        ttcq_QueueError = _ttc_queue_pointer_push_back(Queue, Data);
        if (ttcq_QueueError == tqe_OK) // push successfull: increase value
            ttc_semaphore_give( &(Queue->Amount), 1);
    }
    else { // queue is currently locked by a task: store pending pointer in extra isr-storage
        if (Queue->AmountPendingPushes < TTC_SQB_ISR_SPACE) {
            Queue->PendingPush[Queue->AmountPendingPushes++] = Data;
            ttcq_QueueError = tqe_IsrOperationPending;
        }
        else // no more space left for pending pushes
            ttcq_QueueError = tqe_QueueIsLocked;
    }

    return ttcq_QueueError;
}
    ttc_queue_error_e ttc_queue_pointer_pull_front(ttc_queue_pointers_t* Queue, void** Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(Data,  ec_InvalidArgument);

    ttc_semaphore_take( &(Queue->Amount), 1, -1, ttc_queue_pointer_pull_front);
    ttc_mutex_lock( &(Queue->Lock) , -1, ttc_queue_pointer_pull_front);
    ttc_queue_error_e Error = _ttc_queue_pointer_pull_front(Queue, Data);
    if (Queue->AmountPendingPushes) { // data from isr is pending to be pushed: complete push
        ttc_task_begin_criticalsection(ttc_queue_byte_pull_front);
        ttc_interrupt_all_disable();
        _ttc_queue_pointer_process_pending(Queue);
        ttc_interrupt_all_enable();
        ttc_task_end_criticalsection();
    }
    ttc_mutex_unlock( &(Queue->Lock) );
    if (Error > tqe_Error) { // could not read data: give back token
        ttc_semaphore_give( &(Queue->Amount), 1);
    }

    return Error;
}
    ttc_queue_error_e ttc_queue_pointer_pull_front_isr(ttc_queue_pointers_t* Queue, void** Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(Data,  ec_InvalidArgument);

    if (ttc_semaphore_available(&(Queue->Amount))) { // queue stores data: try to pull one
        if (! ttc_mutex_is_locked( &(Queue->Lock) ) ) {
            if (Queue->AmountPendingPushes)
                _ttc_queue_pointer_process_pending(Queue);
            ttcq_QueueError = _ttc_queue_pointer_pull_front(Queue, Data);
            if (ttcq_QueueError == tqe_OK) {
                ttc_semaphore_take_isr(&(Queue->Amount), 1, ttc_queue_pointer_pull_front_isr);
            }
        }
        else { // queue is currently locked by a task
            ttcq_QueueError = tqe_QueueIsLocked;
        }
        return ttcq_QueueError;
    }
    return tqe_QueueIsEmpty;
}
    ttc_queue_error_e ttc_queue_pointer_try_pull_front(ttc_queue_pointers_t* Queue, void** Data) {
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(Data,  ec_InvalidArgument);

    ttc_queue_error_e Error = ttc_semaphore_try_take( &(Queue->Amount), 1, ttc_queue_pointer_try_pull_front);
    if (Error == tqe_OK) {
        ttc_mutex_lock( &(Queue->Lock) , -1, ttc_queue_pointer_try_pull_front);
        Error = _ttc_queue_pointer_pull_front(Queue, Data);
        ttc_mutex_unlock( &(Queue->Lock) );
        if (Error > tqe_Error) { // could not read data: give back token
            ttc_semaphore_give( &(Queue->Amount), 1);
        }
    }

    return Error;
}
