#ifndef ttc_queue_H
#define ttc_queue_H

/*{ ttc_queue.h ***********************************************
 * 
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported schedulers: FreeRTOS
 * 
}*/
/*{ Details of Queue-Implementation

  ttc_queue provides two main types of queues:
  1) Generic Queues
     These are just a shallow interface to an external supplied queue implementation.
     The generic queue interface was developed on top of FreeRTOS queues. Other queue implementations may follow.
     Depending on their individual implementation, generic queues can consume lots of ram.

  2) Simple Queues
     The most basic datatypes being placed in queues are single bytes and pointers. In ttc_queue, specialized
     implementations for queueing these simple datatypes are provided. The implementations fully support multitasking
     and interrupt service routines with small extra ram usage. The support for interrupt service routines is special.
     When an _isr() function should operate on a simple queue, it can happen that this queue is currently in use by a
     task. This is indicated by Lock==1. In such a case, the isr has a problem. It cannot wait until the task finishes.
     For this situation, an extra space is reserved in the queue buffer for data pending to be pushed.
  }*/

//{ includes

#include "ttc_basic.h"
#include "ttc_memory.h"
#include "ttc_mutex.h"
#include "ttc_semaphore.h"

#ifdef EXTENSION_300_scheduler_free_rtos
  #include <string.h>
  #include "stddef.h"
  #include "FreeRTOS.h"
  #include "queue.h"
  #include "semphr.h"

  typedef void* ttc_QueueHandle;
#else
  typedef void* ttc_QueueHandle;
#endif

typedef enum ttc_queue_error_E {      // return type of most functions in ttc_queue
    tqe_OK,                           // no error
    tqe_HigherPriorityTaskWoken,      // an _isr()-call has unblocked a higher priority task to unblock. Call ttc_task_yield() before exiting your ISR!
    tqe_IsrOperationPending,          // operation is stored for later processing. Call again to complete _isr() operation

    tqe_Error,
    // Error Conditions below
    tqe_TimeOut,                      // an operation has timed out
    tqe_NotImplemented,               // required function not implemented for current architecture
    tqe_QueuePushFailed,              // pushing data to a queue has failed
    tqe_QueueIsEmpty,                 // pulling data from a queue has failed because its empty
    tqe_QueueIsLocked,                // data could not be pulled/pushed, because queue is locked by another task
    tqe_UNKNOWN
} ttc_queue_error_e;

//}includes
//{ Default settings (change via makefile) ******************************

#ifndef TTC_ASSERT_QUEUES
  #define TTC_ASSERT_QUEUES 1
#endif

//}Default settings
//{ Types & Structures ***************************************************


#if TTC_ASSERT_QUEUES == 1 // use Assert()s in Queue code (somewhat slower but alot easier to debug)
  #define Assert_QUEUES(Condition, ErrorCode) Assert(Condition, ErrorCode)
#else  // us no Assert()s in USART code (somewhat smaller + faster, but crashes are hard to debug)
  #define Assert_QUEUES(Condition, ErrorCode)
#endif

// extra memory space in queue reserved as buffer for ISR-operations
#define TTC_SQB_ISR_SPACE 3 // check ttc_queue_bytes_t to get proper aligned end of PendingPush[]!

typedef struct ttc_queue_waiting_task_s {
    TaskHandle_t Task;                        // handle of task waiting on queue
    struct ttc_queue_waiting_task_s* Next;
} ttc_queue_waiting_task_t;

typedef struct { // data of a simple queue for 8-bit values
    ttc_mutex_smart_t Lock;                    // protects queue from concurrent accesses
    ttc_semaphore_smart_t Amount;              // amount of items stored in queue (can be used to wait on queue)
    ttc_queue_waiting_task_t* FirstTaskWaiting2Push; // list of sleeping tasks waiting to push to this queue
    ttc_queue_waiting_task_t* LastTaskWaiting2Push;  // last list entry (for fast list append)
    Base_t Start;                              // index of first valid element in Data[]
    Base_t End;                                // index of last valid element in Data[]
    Base_t Size;                               // size of Data[]
#if TTC_SQB_ISR_SPACE > 0
    u8_t AmountPendingPushes;                  // amount of ISR-pushs still pending to be processed
    u8_t PendingPush[TTC_SQB_ISR_SPACE];       // data pending to be pushed to queue
    //
    // Note: u8_t AmountPendingPushes requires proper value of TTC_SQB_ISR_SPACE to ensure correct alignment!
#ifdef TTC_REGRESSION
    Base_t AmountPendingTotal;                 // total amount of pending operations in this queue so far
    Base_t AmountTaskWaitingTotal;             // total amount of tasks being put to sleep while waiting to push
#endif
#endif
    u8_t* Data;                                // stores 8-bit values in circular buffer
} ttc_queue_bytes_t;


/*???
typedef struct {
  // Mutex protecting queue structure
  unsigned char Lock;

  // semaphore storing queue size
  unsigned char Amount;

  // index of first valid element in Data[]
  unsigned short Start;

  // index of last valid element in Data[]
  unsigned short End;

  // size of Data[] - 1
  unsigned short Size;

  // amount of pending pushes from ISR
  unsigned char AmountPending;

  // data pending to be pushed to queue
  unsigned char Pending[TTC_SQB_ISR_SPACE];

  // stores 8-bit values in circular buffer
  u8_t Data[];
} queue_bytes_t;
*/

typedef struct { // data of a simple queue for pointer values
    ttc_mutex_smart_t Lock;                // protects queue from concurrent accesses
    ttc_semaphore_smart_t Amount;          // amount of items stored in queue (can be used to wait on queue)
    Base_t Start;                          // index of first valid element in Data[]
    Base_t End;                            // index of last valid element in Data[]
    Base_t Size;                           // size of Data[]
#if TTC_SQB_ISR_SPACE > 0
    Base_t AmountPendingPushes;            // amount of ISR-pushs still pending to be processed
    void*  PendingPush[TTC_SQB_ISR_SPACE]; // data pending to be pushed to queue
#endif
    void** Data;                  // stores pointers in circular buffer
} ttc_queue_pointers_t;

//}
//{ includes 2

#include "ttc_interrupt.h"

//}includes 2
//{ Function prototypes **************************************************

/** creates a new generic queue  (Allocates required memory block from heap).
 * @param AmountOfEntries  maximum amount of entries that the new queue can store
 * @param SizeOfEntries    size of each entry in queue
 * @return                 =! NULL: queue has been created successfully (use handle to operate on queue)
 */
ttc_QueueHandle ttc_queue_create(Base_t AmountOfEntries, Base_t SizeOfEntries);

/** initializes given memory for use as simple byte queue.
 * Simple queues are faster and use less memory than general queues.
 * Simple queues are available even without a multitasking scheduler.
 *
 * Note: use of this function can be dangerous, because the valid size of *Storage[] cannot be verified!
 *
 * Restrictions are:
 * - only one task is alowed to push data into the queue
 * - only one task is alowed to pop data from the queue
 * - fixed item size
 *
 * @param Queue            address of static or dynamically allocated memory space to use for queue header
 * @param Storage          memory space to use for queue data (must store AmountOfEntries bytes!)
 * @param AmountOfEntries  Add 1 for queu-management! Queue is created with a fixed size. It cannot be resized later.
 * @return pointer to struct containing all necessary data of this queue
 */
void ttc_queue_byte_init( ttc_queue_bytes_t* Queue, u8_t* Storage, u16_t AmountOfEntries);

/** adds given entry ad end of given queue
 *
 * Note: This function requires a running task scheduler!
 *
 * @param Queue    handle of already created queue
 * @param Entry    pointer to data to be copied into new queue entry
 * @param TimeOut  amount of system ticks to wait for queue to become available
 * @return         == tqe_OK: Entry has been successfully added to Queue within TimeOut
 */
ttc_queue_error_e ttc_queue_push_back(ttc_QueueHandle Queue, const void* Entry, Base_t TimeOut);

/** adds given entry ad end of given queue (does never block)
 * Note: Must be called from ISR or with disabled scheduler! (-> ttc_task_begin_criticalsection )
 *
 * @param Queue    handle of already created queue
 * @param Entry    pointer to data to be copied into new queue entry
 * @param TimeOut  amount of system ticks to wait for queue to become available
 * @return         == tqe_OK: Entry has been successfully added to Queue within TimeOut
 *                 == tqe_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 */
ttc_queue_error_e ttc_queue_push_back_isr(ttc_QueueHandle Queue, const void* Entry);

/** removes first entry from given queue and copies its data into given buffer
 *
 * Note: This function requires a running task scheduler!
 *
 * @param Queue  handle of already created queue
 * @param Entry  pointer to buffer where content of first entry should be copied to
 * @param TimeOut  amount of system ticks to wait for queue to provide data
 * @return         == tqe_OK: Entry has been successfully added to Queue within TimeOut
 */
ttc_queue_error_e ttc_queue_pop_front(ttc_QueueHandle Queue, void* Entry, Base_t TimeOut);

/** removes first entry from given queue and copies its data into given buffer (does never block)
 * @param Queue  handle of already created queue
 * @param Entry  pointer to buffer where content of first entry should be copied to
 * @return         == tqe_OK: Entry has been successfully added to Queue within TimeOut
 */
ttc_queue_error_e ttc_queue_pop_front_isr(ttc_QueueHandle Queue, void* Entry);

/** provides amount of entries currently waiting in given queue
 *
 * Note: This function requires a running task scheduler!
 *
 * @param Queue  handle of already created queue
 * @return       current amount of occupied entries
 */
Base_t ttc_queue_amount_waiting(ttc_QueueHandle Queue);

/** Creates a simple queue for 8-bit data  (Allocates required memory block from heap).
 * Simple queues are faster and use less memory than general queues.
 * Simple queues are available even without a multitasking scheduler.
 * Functions on simple queues will never block.
 *
 * Restrictions are:
 * - only one task is alowed to push data into the queue
 * - only one task is alowed to pop data from the queue
 * - fixed item size
 *
 * @param AmountOfEntries  Queue is created with a fixed size. It cannot be resized later.
 * @return pointer to struct containing all necessary data of this queue
 */
ttc_queue_bytes_t* ttc_queue_byte_create(u16_t AmountOfEntries);

/** returns amount of entries currently stored in given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: Returned size can be corrupt if items are added/ removed by another task!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         Byte to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
Base_t ttc_queue_byte_get_amount(ttc_queue_bytes_t* Queue);

/** adds given byte at end of given queue
 *
 * Note: This function is may be called from different tasks on same queue.
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         Byte to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
void ttc_queue_byte_push_back(ttc_queue_bytes_t* Queue, u8_t Data);

/** adds given byte at end of given queue (never blocks)
 *
 * Note: This function may only be called if interrupts are disabled (e.g. from an interrupt service routine)
 * Note: If queue is locked for pushing, data will be stored in a small extra buffer. If this buffer expires, data is not stored!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         Byte to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 *                     == tqe_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 *                     >= tqe_Error: Error condition
 */
ttc_queue_error_e ttc_queue_byte_push_back_isr(ttc_queue_bytes_t* Queue, u8_t Data);

/** adds given byte at end of given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: It is safe to have one single task call push and another single task call pull on the same queue.
 * Note: It is not safe to have >1 task push/ pull on same queue!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         Byte to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
ttc_queue_error_e _ttc_queue_byte_try_push_back(ttc_queue_bytes_t* Queue, u8_t Data);

/** reads one byte from given queue
 *
 * Note: This function may be called from different tasks
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         byte will be written to this address (no write in case of error)
 * @return             == tqe_OK: Entry has been successfully pulled
 */
ttc_queue_error_e ttc_queue_byte_pull_front(ttc_queue_bytes_t* Queue, u8_t* Data);

/** reads one byte from given queue (call from interrupt service routine)
 *
 * Note: This function may be called from different tasks
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         byte will be written to this address (no write in case of error)
 * @return             == tqe_OK: Entry has been successfully pulled
 *                     == tqe_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 *                     >= tqe_Error: Error condition
 */
ttc_queue_error_e ttc_queue_byte_pull_front_isr(ttc_queue_bytes_t* Queue, u8_t* Data);
/** reads one byte from given queue (call from interrupt service routine)
 *
 * Note: This function may be called from different tasks
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         byte will be written to this address (no write in case of error)
 * @return             == tqe_OK: Entry has been successfully pulled
 *                     == tqe_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 *                     >= tqe_Error: Error condition
 */
ttc_queue_error_e ttc_queue_byte_try_pull_front(ttc_queue_bytes_t* Queue, u8_t* Data);
/** processes all pending isr operations on given byte queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: It is safe to have one single task call push and another single task call pull on the same queue.
 * Note: It is not safe to have >1 task push/ pull on same queue!
 * Note: This function is private and should not be called from outside!
 *
 * @param Queue        datastructure of already created queue
 */
void _ttc_queue_byte_process_pending(ttc_queue_bytes_t* Queue);

/** Creates a simple queue to store pointers (Allocates required memory block from heap).
 * Simple queues are faster and use less memory than general queues.
 * Simple queues are available even without a multitasking scheduler.
 *
 * Restrictions are:
 * - only one task is alowed to push data into the queue
 * - only one task is alowed to pop data from the queue
 * - fixed item size
 *
 * @param AmountOfEntries  Queue is created with a fixed size. It cannot be resized later.
 * @return pointer to struct containing all necessary data of this queue
 */
ttc_queue_pointers_t* ttc_queue_pointer_create(u16_t AmountOfEntries);

/** initializes given memory for use as simple pointer queue.
 * Simple queues are faster and use less memory than general queues.
 * Simple queues are available even without a multitasking scheduler.
 *
 * Note: use of this function can be dangerous, because the valid size of *Storage[] cannot be verified!
 *
 * Restrictions are:
 * - only one task is alowed to push data into the queue
 * - only one task is alowed to pop data from the queue
 * - fixed item size
 *
 * @param Queue            address of static or dynamically allocated memory space to use
 * @param Storage          memory space to use for queue data (must store AmountOfEntries+1 * sizeof(Base_t) bytes!)
 * @param AmountOfEntries  Add 1 for queu-management! Queue is created with a fixed size. It cannot be resized later.
 * @return pointer to struct containing all necessary data of this queue
 */
void ttc_queue_pointer_init(ttc_queue_pointers_t* Queue, void** Storage, u16_t AmountOfEntries);

/** returns amount of entries currently stored in given queue
 *
 * Note: This function is not protected against concurrent calls from different tasks!
 * Note: Returned size can be corrupt if items are added/ removed by another task!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         Byte to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
Base_t ttc_queue_pointer_get_amount(ttc_queue_pointers_t* Queue);

/** adds given byte at end of given queue
 *
 * Note: It is safe to have one single task call push and another single task call pull on the same queue.
 * Note: It is not safe to have >1 task push/ pull on same queue!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         pointer address to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
ttc_queue_error_e ttc_queue_pointer_push_back(ttc_queue_pointers_t* Queue, void* Data);

/** adds given byte at end of given queue (never blocks)
 *
 * Note: This function may only be called if interrupts are disabled (e.g. from an interrupt service routine)
 * Note: If queue is locked for pushing, data will be stored in a small extra buffer. If this buffer expires, data is not stored!
 *
 * @param QueueHandle  handle of already created queue
 * @param Data         pointer address to append to queue
 * @return             == tqe_OK: Entry has been successfully added
 */
ttc_queue_error_e ttc_queue_pointer_push_back_isr(ttc_queue_pointers_t* Queue, void* Data);

/** reads one byte from given queue
 *
 * Note: This function can be used from different tasks
 *
 * @param QueueHandle  handle of already created queue
 * @param Pointer      pointer address will be written to this address (NULL if queue is empty)
 * @return             == tqe_OK: Entry has been successfully pulled
 */
ttc_queue_error_e ttc_queue_pointer_pull_front(ttc_queue_pointers_t* Queue, void** Data);

/** tries to read one byte from given queue (will never block)
 *
 * Note: This function can be used from different tasks
 *
 * @param QueueHandle  handle of already created queue
 * @param Pointer      pointer address will be written to this address (NULL if queue is empty)
 * @return             == tqe_OK: Entry has been successfully pulled
 */
ttc_queue_error_e ttc_queue_pointer_try_pull_front(ttc_queue_pointers_t* Queue, void** Data);

/** reads one byte from given queue(never blocks)
 *
 * Note: This function may only be called if interrupts are disabled (e.g. from an interrupt service routine)
 *
 * @param QueueHandle  handle of already created queue
 * @param Pointer      pointer address will be written to this address (NULL if queue is empty)
 * @return             == tqe_OK: Entry has been successfully pulled
 */
ttc_queue_error_e ttc_queue_pointer_pull_front_isr(ttc_queue_pointers_t* Queue, void** Data);

//}Function prototypes

#endif
