#ifndef ttc_semaphore_H
#define ttc_semaphore_H

/** ttc_semaphore.h ***********************************************{
 * 
 * Written by Gregor Rebel 2013
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported architecture/ schedulers: CortexM3 (cm3), FreeRTOS
 * 
 * Why using ttc_semaphore.c instead of cm3_semaphore.c or FreeRTOS-functions directly?
 * (1) it is more portable
 *     You may change to another multitasking scheduler once it is supported by The ToolChain
 * (2) extra safety checks
 *     DeadAmounts are a common issue in multitasking applications and sometimes hard to debug.
 *     + ttc_semaphore allows to turn all endless Amounts into a timed one that will assert if expired.
 *       -> ttc_semaphore_types.h/TTC_SEMAPHORE_TIMEOUT_USECS
 *     + ttc_semaphore provides smart semaphores that can record history of operations and
 *       run as fast as normal semaphores if disabled.
 *       -> ttc_semaphore_types.h/TTC_SMART_SEMAPHORES
 *
 * Smart semaphores store the owner of a Amount.
 * Whenever a semaphore stays Amounted because of a programming error, then the last Amounting
 * function can be identified by taking a look into the smart Amount structure.
}*/

//{ includes

#include "ttc_basic.h"

#ifdef EXTENSION_500_ttc_semaphore_cortexm3
#  include "cm3_semaphore.h"
#endif

#ifdef EXTENSION_500_ttc_semaphore_freertos
#  include "freertos_semaphore.h"
#endif

#include "ttc_semaphore_types.h"
#ifdef EXTENSION_500_ttc_task
#  include "ttc_task.h"
#endif

//}includes
//{ Types & Structures ***************************************************

/** Types to be defined by low-level driver ******************************
 *
 * ttc_semaphore_t  datatype storing a semaphore
 * E.g.: typedef u8_t ttc_semaphore_t
 *
 */


//}
#include "ttc_memory.h"

//{ Function prototypes **************************************************
//
// For each function, a DRIVER_() pendant must be defined by low-level driver to provide an implementation.


/** creates a new semaphore with initial value 0
 *
 * @return                 =! NULL: semaphore has been created successfully (use handle to operate on semaphore)
 */
ttc_semaphore_smart_t* ttc_semaphore_create();
#if TTC_SMART_SEMAPHORES == 1
#  define ttc_semaphore_create() _ttc_semaphore_smart_create()
#else
#  define ttc_semaphore_create() _ttc_semaphore_create()
#endif
ttc_semaphore_smart_t* _ttc_semaphore_smart_create();
ttc_semaphore_t* _ttc_semaphore_create();
#ifndef DRIVER_ttc_semaphore_create
#error Missing low-level implementation! // check description above for function to be implemented
#endif


/** initializes given memory as a new empty semaphore
 *
 * @return                 =! NULL: semaphore has been created successfully (use handle to operate on semaphore)
 */
void ttc_semaphore_init(ttc_semaphore_smart_t* Semaphore);
#if TTC_SMART_SEMAPHORES == 1
#  define ttc_semaphore_init(SmartSemaphore) _ttc_semaphore_smart_init(SmartSemaphore)
#else
#  define ttc_semaphore_init(Semaphore)      _ttc_semaphore_init(Semaphore)
#endif
void _ttc_semaphore_smart_init(ttc_semaphore_smart_t* Semaphore);
void _ttc_semaphore_init(ttc_semaphore_t* Semaphore);
#ifndef DRIVER_ttc_semaphore_init
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Checks if given smart semaphore can provide given amount
 *
 * Note: Calling this function makes only sense from interrupt service routine
 *
 * @param Semaphore  handle of already created semaphore
 * @return         == tsme_OK: Given semaphore can provide given Amount (only valid if no other task or ISR has taken them meanwhile)
 */
BOOL ttc_semaphore_can_provide(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount);
#if TTC_SMART_SEMAPHORES == 1
#  define ttc_semaphore_can_provide(Semaphore, Amount) _ttc_semaphore_smart_can_provide(Semaphore, Amount)
#else
#  define ttc_semaphore_can_provide(Semaphore, Amount) _ttc_semaphore_can_provide(Semaphore, Amount)
#endif
BOOL _ttc_semaphore_smart_can_provide(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount);
BOOL _ttc_semaphore_can_provide(ttc_semaphore_t* Semaphore, Base_t Amount);
#ifndef DRIVER_ttc_semaphore_can_provide
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Tries to take given amount from a semaphore within timeout
 *
 * @param Semaphore  handle of already created semaphore
 * @param TimeOut    amount of system ticks to wait for Amount
 * @param Owner      reference to function which is calling this function
 * @return           == tsme_OK: Amount has been successfully obtained within TimeOut
 */
ttc_semaphore_error_e ttc_semaphore_take(ttc_semaphore_smart_t* Semaphore, Base_t Amount, Base_t TimeOut, void (*Owner)() );
#if TTC_SMART_SEMAPHORES == 1
#  define ttc_semaphore_take(Semaphore, Amount, TimeOut, Owner) _ttc_semaphore_smart_take(Semaphore, Amount, TimeOut, (void (*)()) Owner)
#else
#  define ttc_semaphore_take(Semaphore, Amount, TimeOut, Owner) _ttc_semaphore_take(Semaphore, Amount, TimeOut)
#endif
ttc_semaphore_error_e _ttc_semaphore_smart_take(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount, Base_t TimeOut, void (*Owner)() );
ttc_semaphore_error_e _ttc_semaphore_take(ttc_semaphore_t* Semaphore, Base_t Amount, Base_t TimeOut);
#ifndef DRIVER_ttc_semaphore_take
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Tries to take given amount from semaphore; this function will never block and must only be called with disabled interrupts!
 *
 * @param Semaphore  handle of already created semaphore
 * @param Amount     amount of tokens to give/ take to/from semaphore
 * @param Owner      reference to function which is calling this function
 * @return           == tsme_OK: Given Amount has been successfully taken from semaphore
 *                   == tsme_WasAmountedByTask: - a task of higher priority has been interrupted => call ttc_task_yield() at end of your interrupt service routine!
 *                   >= tsme_Error: Error condition
 */
ttc_semaphore_error_e ttc_semaphore_try_take(ttc_semaphore_smart_t* Semaphore, Base_t Amount, void (*Owner)() );
#if TTC_SMART_SEMAPHORES == 1
#  define ttc_semaphore_try_take(Semaphore, Amount, Owner) _ttc_semaphore_smart_try_take(Semaphore, Amount, (void (*)()) Owner)
#else
#  define ttc_semaphore_try_take(Semaphore, Amount, Owner) _ttc_semaphore_try_take(Semaphore, Amount)
#endif
ttc_semaphore_error_e _ttc_semaphore_smart_try_take(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount, void (*Owner)() );
ttc_semaphore_error_e _ttc_semaphore_try_take(ttc_semaphore_t* Semaphore, Base_t Amount);
// #ifndef DRIVER_ttc_semaphore_try_take
// #error Missing low-level implementation! // check description above for function to be implemented
// #endif


/** Tries to take given amount from semaphore; this function will never block and must only be called with disabled interrupts!
 *
 * @param Semaphore  handle of already created semaphore
 * @param Amount     amount of tokens to give/ take to/from semaphore
 * @param Owner      reference to function which is calling this function
 * @return           == tsme_OK: Given Amount has been successfully taken from semaphore
 *                   == tsme_WasAmountedByTask: - a task of higher priority has been interrupted => call ttc_task_yield() at end of your interrupt service routine!
 *                   >= tsme_Error: Error condition
 */
ttc_semaphore_error_e ttc_semaphore_take_isr(ttc_semaphore_smart_t* Semaphore, Base_t Amount, void (*Owner)() );
#if TTC_SMART_SEMAPHORES == 1
#  define ttc_semaphore_take_isr(Semaphore, Amount, Owner) _ttc_semaphore_smart_take_isr(Semaphore, Amount, (void (*)()) Owner)
#else
#  define ttc_semaphore_take_isr(Semaphore, Amount, Owner) _ttc_semaphore_take_isr(Semaphore, Amount)
#endif
ttc_semaphore_error_e _ttc_semaphore_smart_take_isr(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount, void (*Owner)() );
ttc_semaphore_error_e _ttc_semaphore_take_isr(ttc_semaphore_t* Semaphore, Base_t Amount);
#ifndef DRIVER_ttc_semaphore_take_isr
#error Missing low-level implementation! // check description above for function to be implemented
#endif


/** Gives amount of tokens back to semaphore
 *
 * @param Semaphore  handle of already created smart semaphore
 * @param Amount     amount of tokens to give/ take to/from semaphore
 * @return       == tsme_OK: Amount has been given back successfully
 */
ttc_semaphore_error_e ttc_semaphore_give(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount);
#if TTC_SMART_SEMAPHORES == 1
#define ttc_semaphore_give(Semaphore, Amount) _ttc_semaphore_smart_give(Semaphore, Amount)
#else
#define ttc_semaphore_give(Semaphore, Amount) _ttc_semaphore_give(Semaphore, Amount)
#endif
ttc_semaphore_error_e _ttc_semaphore_smart_give(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount);
ttc_semaphore_error_e _ttc_semaphore_give(ttc_semaphore_t* Semaphore, Base_t Amount);
#ifndef DRIVER_ttc_semaphore_give
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Gives amount of tokens back to semaphore (does never block)
 *
 * @param Semaphore  handle of already created semaphore
 * @return         == tsme_OK: Amount has been given back successfully
 *                 == tsme_WasAmountedByTask: call task ttc_task_yield_isr() before exiting your ISR!
 */
ttc_semaphore_error_e ttc_semaphore_give_isr(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount);
#if TTC_SMART_SEMAPHORES == 1
#  define ttc_semaphore_give_isr(Semaphore, Amount) _ttc_semaphore_smart_give_isr(Semaphore, Amount)
#else
#  define ttc_semaphore_give_isr(Semaphore, Amount) _ttc_semaphore_give_isr(Semaphore, Amount)
#endif
ttc_semaphore_error_e _ttc_semaphore_smart_give_isr(ttc_semaphore_smart_t* SmartSemaphore, Base_t Amount);
ttc_semaphore_error_e _ttc_semaphore_give_isr(ttc_semaphore_t* Semaphore, Base_t Amount);
#ifndef DRIVER_ttc_semaphore_give_isr
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Returns current amount of tokens available in given Semaphore
 *
 * @param Semaphore  handle of already created semaphore
 * @return           current value stored in given semaphore
 */
Base_t ttc_semaphore_available(ttc_semaphore_smart_t* SmartSemaphore);
#if TTC_SMART_SEMAPHORES == 1
#  define ttc_semaphore_available(Semaphore) _ttc_semaphore_smart_available(Semaphore)
#else
#  define ttc_semaphore_available(Semaphore) DRIVER_ttc_semaphore_available(&((Semaphore)->Value))
#endif
Base_t _ttc_semaphore_smart_available(ttc_semaphore_smart_t* SmartSemaphore);
#ifndef DRIVER_ttc_semaphore_available
#error Missing low-level implementation! // check description above for function to be implemented
#endif

//}Function prototypes

#endif
