#ifndef ttc_semaphore_types_H
#define ttc_semaphore_types_H

/*{ ttc_semaphore_types.h ***********************************************
 * 
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported architectures/ schedulers: CortexM3, FreeRTOS
 * 
}*/

//{ includes

#include "ttc_basic.h"
#include "ttc_task_types.h"

#ifdef EXTENSION_500_ttc_semaphore_cortexm3
#  include "cm3_semaphore_types.h"
#endif

#ifdef EXTENSION_500_ttc_semaphore_freertos
#  include "freertos_semaphore_types.h"
#endif


typedef enum ttc_semaphore_error_E {  // return type of most functions in ttc_semaphore
    tsme_OK,                           // no error
    tsme_WasAmountedByTask,            // an _isr()-call occured while Semaphore was Amounted by a task

    tsme_Error,
    // Error Conditions below
    tsme_TimeOut,                      // an operation has timed out
    tsme_NotImplemented,               // required function not implemented for current architecture
    tsme_SemaphoreExpired,             // tried to take from semaphore that cannot provide given amount
    tsme_SemaphoreFull,                // tried to give to semaphore that has reached its limit (if supported)
    tsme_SemaphoreInterrupted,         // semaphore operation was interrupted by another task (retry your action)
    tsme_UNKNOWN
} ttc_semaphore_error_e;

//}includes
//{ Default settings (change via makefile) ******************************


#ifndef TTC_SEMAPHORE_TIMEOUT_USECS
  //#define TTC_SEMAPHORE_TIMEOUT_USECS 1000000 // >0: activate timeout-assert for endless semaphores (TimeOut==-1) in usecs
  #define TTC_SEMAPHORE_TIMEOUT_USECS 0    // no forced timeout
#endif

// enable use of Assert() for semaphores
#ifndef TTC_ASSERT_SEMAPHORES
  #define TTC_ASSERT_SEMAPHORES 1
#endif

//}Default settings
//{ Types & Structures ***************************************************

#if TTC_ASSERT_SEMAPHORES == 1 // use Assert()s in Queue code (somewhat slower but alot easier to debug)
  #define Assert_SEMAPHORES(Condition, ErrorCode) Assert(Condition, ErrorCode)
#else  // us no Assert()s in USART code (somewhat smaller + faster, but crashes are hard to debug)
  #define Assert_SEMAPHORES(Condition, ErrorCode)
#endif

// enable use of smart semaphores
#ifndef TTC_SMART_SEMAPHORES
  #define TTC_SMART_SEMAPHORES 1
#endif

// define amount
#if TTC_SMART_SEMAPHORES == 1

  // set amount of entries to be stored in history of smart semaphore usage (debugging only)
  #ifndef TTC_SMART_SEMAPHORE_HISTORY_SIZE
    #define TTC_SMART_SEMAPHORE_HISTORY_SIZE 0
  #endif

#endif

/** tasks waiting for a semaphore create a single linked list of task-handles on their stacks */
typedef struct ttc_semaphore_waiting_task_s {
    TaskHandle_t Task;                        // handle of task waiting on semaphore
    struct ttc_semaphore_waiting_task_s* Next;
} ttc_semaphore_waiting_task_t;

/** data used to manage a single semaphore */
typedef struct {
  DRIVER_ttc_semaphore_t        Value;            // low-level data storing value (defined in low-level driver: cm3_semaphore.h, ...)
  ttc_semaphore_waiting_task_t* FirstWaitingTask; // points to list of tasks waiting on this semaphore
  ttc_semaphore_waiting_task_t* LastWaitingTask;  // points to last list entry (for fast list append)
} ttc_semaphore_t;

// debuggable semaphore that can store its owner
// This comes handy if you wanna know, who has taken a semaphore before.
#if TTC_SMART_SEMAPHORES == 1
  typedef struct ttc_semaphore_smart_s {
    ttc_semaphore_t Semaphore __attribute__((__aligned__(4)));

    // function pointer given to Amount() function
    void (*Owner)();

#ifdef TTC_REGRESSION
    // task info of owner
    ttc_task_info_t* OwnerTask;
#endif
  } ttc_semaphore_smart_t;
#else
  typedef ttc_semaphore_t ttc_semaphore_smart_t;
#endif

//}

/** called while waiting for semaphore Amount. Place your breakpoint inside to debug hanging Amounts
  */
void ttc_semaphore_wait(ttc_semaphore_t* Semaphore);



#endif
