#ifndef ttc_task_H
#define ttc_task_H

/*{ ttc_task.h ***********************************************
 
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for multitasking.
 * 
 * Currently implemented architectures: stm32 + FreeRTOS
 * 
}*/

//{ includes

#include "ttc_basic.h"
#include "ttc_gpio.h"
#include "ttc_memory.h"
#include "ttc_string.h"
#include "ttc_task_types.h"

#ifdef EXTENSION_300_scheduler_free_rtos
  #include "scheduler/freertos_task.h"
#endif

//}includes

//{ Function prototypes **************************************************

/** produces software PWM usable to calibrate uSleep()
 * Idea: Connect oscilloscope to pin PA1 (pin 15 on STM32F103) an check for correct period (1,1ms) and duty cycle.
 *        Adjust uSleep() accordingly, recompile, reflash and retest.
 */
void ttc_task_calibrate_usleep();

/** Set task to sleep for some amount of time
 *
 * Note: Correct timing is not guaranteed if task scheduler is not running!
 *
 * @param Delay    delay time (msecs)
*/
void ttc_task_msleep(Base_t Delay);
void _ttc_task_msleep(Base_t Delay);
#ifdef DRIVER_ttc_task_msleep
#  define ttc_task_msleep(Delay)  DRIVER_ttc_task_msleep(Delay)
#else
#  define ttc_task_msleep(Delay)  _ttc_task_msleep(Delay)
#endif
#define mSleep(Delay) ttc_task_msleep(Delay)

/** set task to sleep for some amount of time
 *
 * Note: This function may only be called after ttc_task_start_scheduler()!
 *
 * @param Delay    delay time (usecs)
 */
void ttc_task_usleep(Base_t Delay);
#ifdef DRIVER_ttc_task_usleep
#define ttc_task_usleep(Delay) DRIVER_ttc_task_usleep(Delay)
#else
#define ttc_task_usleep(Delay) _ttc_task_fallback_usleep(Delay)
#endif
#define uSleep(Delay) ttc_task_usleep(Delay)
void _ttc_task_fallback_usleep(Base_t Delay);

/** returns smallest supported delay of ttc_task_usleep() for current configuration
 * @return smallest supported delay time (usecs)
 */
#ifdef DRIVER_ttc_task_min_delay
#define ttc_task_min_delay DRIVER_ttc_task_min_delay
#else
#define ttc_task_min_delay 1
#endif
#define MinimumDelay ttc_task_min_delay


/** spawns a function as separate task
 * @param TaskFunction  function to start as task (must never return!)
 * @param TaskName      used for debugging purposes only
 *                      Note: TaskName string will be copied if in writable memory. Use constant strings to safe RAM.
 * @param StackSize     each task has its own size; choose wisely
 * @param Argument      passed as argument to TaskFunction()
 * @param Priority      tasks with higher priority get more cpu time
 * @param Info          !=NULL: Info configuration of current task control block is copied into given variable
 * @return              == TRUE: task was created successfully
 */
BOOL ttc_task_create( void TaskFunction(void*), const char* TaskName, u16_t StackSize, void* Argument, u8_t Priority, ttc_task_info_t** Handle);
BOOL DRIVER_ttc_task_create( void TaskFunction(void*), u16_t StackSize, void* Argument, u8_t Priority, ttc_task_info_t* TaskInfo);
#ifndef DRIVER_ttc_task_create
#warning Missing low-level implementation for DRIVER_ttc_task_create()
#endif

/** Switches scheduler to next process providing a short, undetermined sleep
 *
 * Some applications require a busy waiting for resources like:
 *   while (...);
 * During this loop, the cpu is completely busy with this loop. This blocks or at least slows down other tasks.
 * Even if this task can be interrupted by the task scheduler, it fills its time slice with checking its condition (...).
 *
 * A more multitask-friendly version is to yield cpu time to other processes while waiting:
 *   while (...) ttc_task_yield();
 * For every checked condition (...), every task gets one slice of cpu time. The processing overhead is minimal.
 *
 */
void ttc_task_yield();
#ifdef DRIVER_ttc_task_yield
#define ttc_task_yield() DRIVER_ttc_task_yield()
#else
#define ttc_task_yield()
#endif
void _ttc_task_yield();

/** similar to ttc_task_yield() for use in interrupt service routines only
  */
void ttc_task_yield_isr();
#ifndef DRIVER_ttc_task_yield_isr
#define ttc_task_yield_isr() DRIVER_ttc_task_yield_isr()
#else
#define ttc_task_yield_isr()
#endif

/** Temporarily disables task scheduler*/
void ttc_task_begin_criticalsection(void (*Owner)());
#ifdef DRIVER_ttc_task_begin_criticalsection
#  ifdef TTC_ASSERT_TASK
#    define ttc_task_begin_criticalsection(OWNER) ttc_task_begin_criticalsection_owner( (void (*)()) OWNER)
#  else
#    define ttc_task_begin_criticalsection(OWNER) DRIVER_ttc_task_begin_criticalsection()
#  endif
#else
#    define ttc_task_begin_criticalsection(OWNER) (void) OWNER
#endif
void ttc_task_begin_criticalsection_owner(void (*Owner)());

/** Reenables task scheduler */
void ttc_task_end_criticalsection();
void ttc_task_end_criticalsection_owner();
#ifdef DRIVER_ttc_task_end_criticalsection
#  define ttc_task_end_criticalsection() ttc_task_end_criticalsection_owner()
#else
#  define ttc_task_end_criticalsection() (void)
#endif

/** returns current system clock value
 *
 * Note: Check configuration of your scheduler to achieve real world usecs! (E.g. FreeRTOSConfig.h:configTICK_RATE_HZ)
 *
 * @return amount of usecs since start of system/ scheduler
 */
Base_t ttc_task_get_tick_count();
#ifdef DRIVER_ttc_task_get_tick_count
#define ttc_task_get_tick_count() DRIVER_ttc_task_get_tick_count()
#else
#define ttc_task_get_tick_count() 0
#endif

/** checks if stack of current task or any other task has been overflowed in the past.
 * @param != 0: handle of task which stack should be inspected; ==0: inspect current task
 * @return minimum seen stack bytes free so far
 */
Base_signed_t ttc_task_check_stack();
#ifndef DRIVER_ttc_task_check_stack
#  define DRIVER_ttc_task_check_stack() 1
#endif

/** starts multitasking scheduler (will not return if scheduler is available)
 */
void ttc_task_start_scheduler();

/** checks if task scheduler is currently running
  *
  * Some functions require a running task scheduler. Use this function
  * with Assert() to avoid system crashes before the scheduler has started.
  *
  * @return  == TRUE: ttc_task_start_scheduler() has been called
  */
BOOL ttc_task_is_scheduler_running();

/** resumes a suspended task
  *
  * Note: Do not call this function from an interrupt service routine!
  *
  * @param Handle task identifier as returned by ttc_task_create
  */
void ttc_task_resume(TaskHandle_t Handle);
#ifdef DRIVER_ttc_task_resume
//#  define ttc_task_resume(Handle)     (Handle) ? DRIVER_ttc_task_resume(Handle) : Assert_Halt_EC(ec_InvalidArgument)
#  define ttc_task_resume(Handle)     DRIVER_ttc_task_resume(Handle)
#else
#  define ttc_task_resume(Handle)
#endif

/** resumes a suspended task from an interrupt service routine
  *
  * @param   Handle task identifier as returned by ttc_task_create
  * @return  == TRUE: you must call ttc_task_yield_isr() at end of your isr!
  */
BOOL ttc_task_resume_isr(TaskHandle_t Handle);
#ifdef DRIVER_ttc_task_resume_isr
//#  define ttc_task_resume_isr(Handle)     (Handle) ? DRIVER_ttc_task_resume_isr(Handle) : Assert_Halt_EC(ec_InvalidArgument)
#  define ttc_task_resume_isr(Handle)     DRIVER_ttc_task_resume_isr(Handle)
#else
#  define ttc_task_resume_isr(Handle)
#endif

/** puts a task to sleep.
  *
  * @param Handle != NULL: task identifier as returned by ttc_task_create
  *               == NULL: put current task to sleep
  */
void ttc_task_suspend(TaskHandle_t Handle);

/** taskhandle of current task
  *
  */
TaskHandle_t ttc_task_current_handle();
#ifdef DRIVER_ttc_task_current_handle
#define ttc_task_current_handle() DRIVER_ttc_task_current_handle()
#else
#define ttc_task_current_handle() NULL
#endif

/** returns configuration info of current task
  *
  * Note: This function is specially designed for online debugging.
  *        To find out more about your current task, you may want to issue this in your gdb session:
  *        (gdb) p *(ttc_task_info(NULL, NULL))
  *
  * Note: shortcuts are defined in ttc_basic.h
  *
  * @param Handle  !=NULL: get info of corresponding task
  *                ==NULL: get info of current task
  * @param Info    !=NULL: Info configuration of current task control block is copied into given variable
  *                ==NULL: a local static buffer will be allocated (do not use from multiple threads!)
  * @return pointer to task info structure
  */
ttc_task_info_t* ttc_task_info(TaskHandle_t Handle);
#ifndef DRIVER_ttc_task_update_info
#define DRIVER_ttc_task_update_info(Handle, Info)
#endif

/** scans list of all task infos and returns entry corresponding to given task handle
  *
  * Note: This function performs a linear search in ttc_task_TaskInfos
  *
  *
  * @param Handle  task handle used by multitasking scheduler
  * @return        !=NULL: task info data of given task handle
  */
ttc_task_info_t* ttc_task_find_info(TaskHandle_t Handle);

/** returns configuration info of all tasks
  *
  * Note: This function is specially designed for online debugging.
  *        To find out more about your current task, you may want to issue this in your gdb session:
  *        (gdb) p *(ttc_task_get_all(NULL))
  *
  * Note: shortcuts are defined in ttc_basic.h
  *
  * @param !=NULL: Info configuration of current task control block is copied into given variable
  *        ==NULL: a local static buffer will be allocated (do not use from multiple threads!)
  * @return pointer to array of task info structures
  */
ttc_task_all_t* ttc_task_get_all(ttc_task_all_t* TaskInfos);
#ifndef DRIVER_ttc_task_get_all
#define DRIVER_ttc_task_get_all(TaskInfos)
#endif

/** returns total amount of currently registered tasks
  *
  * @return amount of tasks currently registered
  */
Base_t ttc_task_get_amount();
#ifndef DRIVER_ttc_task_get_amount
#define DRIVER_ttc_task_get_amount() 1
#endif
#define ttc_task_get_amount() DRIVER_ttc_task_get_amount()


//}Function prototypes

#endif
