/*{ regression_memory_pools.c ************************************************
 
  regression_memory_pools_start() spawns thread taskBrightness()

  taskBrightness()
      Defines brightness of led as Ton / (Ton + Toff)
      spawns thread taskLEDs()
      while (1) {
        while (Ton < 100) { Ton++; Toff--; }
        while (Ton >   0) { Ton--; Toff++; }
      }

  taskLEDs()
      while (1) {
        switch on LED
        wait for Ton
        switch off LED
        wait for Toff
      }

}*/

#include "regression_memory_pools.h"

// read-only configuration of current display (width, height, ...)
ttc_gfx_generic_t* DisplayConfig = NULL;

//{ Function definitions *************************************************
//{ FCTMA-Implementation ****************************************************

#define mutex_create     ttc_mutex_create
#define mutex_lock       ttc_mutex_lock
#define mutex_unlock     ttc_mutex_unlock
#define random(MAX)      ttc_rand() % MAX

// ==1: multitasking is disabled during memory operations
#define AVOID_TASK_SWITCHES 1

// select implementation to use as Dynamic Memory Allocator
typedef enum {
    dma_FCTMA,           // Fast Constant Time Memory Allocator using memory pools + memory headers
    dma_SDMA,            // using implementation from heap/heap_sdma.c
    dma_FreeRTOS         // heap memory allocator as provided by FreeRTOS
} DMA_Algorithm_e;
#define DMA_Algorithm dma_FreeRTOS

// structures

// global variables
static ttc_memory_pool_t* Pool100 = NULL;

// amount of blocks allocated on heap
extern Base_t ttc_memory_Amount_Blocks_Allocated;

// amount of bytes occupied on heap by dynamic allocations
extern Base_t ttc_memory_Bytes_Allocated_Total;

// usable amount of allocated bytes
extern Base_t ttc_memory_Bytes_Allocated_Usable;

//} FCTMA-Implementation
//{ Global Variables *****************************************************

static ttc_memory_pool_t* PoolPrintJobs;
void taskProducer(void* Arg);
void taskConsumer(void* Arg);
u8_t DisplayY = 2;
ttc_list_t BlockList;
Base_t AmountHeapFreeAtStart = 0;

// amount of producer tasks to spawn
#define Amount_Producer 10

// amount of consumer tasks to spawn
#define Amount_Consumer 10

// amount of blocks to allocate in each turn
#define Amount_BlocksPerTurn  20

// size of each memory block
#define Block_Size 127


//} Global Variables

typedef struct {
    // read by task
    Base_t SleepTime_Max;

    // written by task
    Base_t BlocksProcessed;    // total amount of memory blocks processed by this task so far
    Base_t SleepTime_Total;    // total amount of sleeps of this task so far
    Base_t SleepTime_Average;  // = SleepTime_Total / BlocksProcessed
} TaskInfo_t;

TaskInfo_t  ProducerTasks[Amount_Producer + 1];
TaskInfo_t* EndOfProducerTasks = ProducerTasks;
TaskInfo_t  ConsumerTasks[Amount_Consumer + 1];
TaskInfo_t* EndOfConsumerTasks = ConsumerTasks;

void* mallocBlock() {
    ttc_gpio_init(TTC_MEMORY_DEBUG_POOL_ALLOC_PIN, tgm_output_push_pull);
    void* Block = NULL;

    switch (DMA_Algorithm) {
    case dma_FCTMA:
        Block = (void*) ttc_memory_pool_alloc(Pool100);
        break;
    case dma_FreeRTOS:
#ifdef TTC_MEMORY_DEBUG_POOL_ALLOC_PIN
        if (0) ttc_task_begin_criticalsection(mallocBlock);
        //ttc_gpio_set(TTC_MEMORY_DEBUG_POOL_ALLOC_PIN);
#endif
        Block = pvPortMalloc(Block_Size);
#ifdef TTC_MEMORY_DEBUG_POOL_ALLOC_PIN
        //ttc_gpio_clr(TTC_MEMORY_DEBUG_POOL_ALLOC_PIN);
        if (0) ttc_task_end_criticalsection();
#endif
        break;
    case dma_SDMA:
#ifdef TTC_MEMORY_DEBUG_POOL_ALLOC_PIN
        if (1) ttc_task_begin_criticalsection(mallocBlock);
        ttc_gpio_set(TTC_MEMORY_DEBUG_POOL_ALLOC_PIN);
#endif
#ifdef EXTENSION_500_ttc_memory_heap_sdma
        Block = heap_sdma_alloc(Block_Size);
#else
        Assert_Halt_EC(ec_InvalidConfiguration); // heap_sdma must be choosen as memory allocator
#endif
#ifdef TTC_MEMORY_DEBUG_POOL_ALLOC_PIN
        ttc_gpio_clr(TTC_MEMORY_DEBUG_POOL_ALLOC_PIN);
        if (1) ttc_task_end_criticalsection();
#endif
        break;
    default:
        Assert_Halt_EC(ec_NotImplemented);
        break;
    }

    return Block;
}
void releaseBlock(ttc_memory_from_pool_t* Block) {
    ttc_gpio_init(TTC_MEMORY_DEBUG_POOL_RELEASE_PIN, tgm_output_push_pull);

    switch (DMA_Algorithm) {
    case dma_FCTMA:
        ttc_memory_pool_release(Block);
        break;
    case dma_FreeRTOS:
#ifdef TTC_MEMORY_DEBUG_POOL_RELEASE_PIN
        if (0) ttc_task_begin_criticalsection(releaseBlock);
        //ttc_gpio_set(TTC_MEMORY_DEBUG_POOL_RELEASE_PIN);
#endif
        vPortFree(Block);
#ifdef TTC_MEMORY_DEBUG_POOL_RELEASE_PIN
        ttc_gpio_clr(TTC_MEMORY_DEBUG_POOL_RELEASE_PIN);
        if (0) ttc_task_end_criticalsection();
#endif
        break;
    case dma_SDMA:
#ifdef TTC_MEMORY_DEBUG_POOL_RELEASE_PIN
        if (1) ttc_task_begin_criticalsection(releaseBlock);
        ttc_gpio_set(TTC_MEMORY_DEBUG_POOL_RELEASE_PIN);
#endif
#ifdef EXTENSION_500_ttc_memory_heap_sdma
        heap_sdma_free(Block);
#else
        Assert_Halt_EC(ec_InvalidConfiguration); // heap_sdma must be choosen as memory allocator
#endif
#ifdef TTC_MEMORY_DEBUG_POOL_RELEASE_PIN
        ttc_gpio_clr(TTC_MEMORY_DEBUG_POOL_RELEASE_PIN);
        if (1) ttc_task_end_criticalsection();
#endif
        break;
    default:
        Assert_Halt_EC(ec_NotImplemented);
        break;
    }
}
void push2Consumer(void* Block) {

    ttc_list_push_back_single(&BlockList, (ttc_list_item_t*) Block);
}
void* popFromProducer() {
    void* Block;

    Block = (void*) ttc_list_pop_front_single(&BlockList);

    return Block;
}
void regression_memory_pools_start() {

#ifdef TTC_MEMORY_DEBUG_POOL_RELEASE_PIN
#endif
#ifdef TTC_MEMORY_DEBUG_POOL_ALLOC_PIN
#endif

    u8_t MaxDisplay = ttc_gfx_get_max_display_index();
    Assert(MaxDisplay > 0, ec_DeviceNotFound); // no displays defined by board makefile!
    DisplayConfig = ttc_gfx_get_configuration(1);
    Assert(DisplayConfig, ec_DeviceNotFound);

    // set color to use for first screen clear
    ttc_gfx_set_color_bg24(GFX_COLOR24_AQUA);
    ttc_gfx_init(1);

    // initialize gfx for use with printf()
    ttc_gfx_stdout_set(1);
    ttc_string_register_stdout(ttc_gfx_stdout_print_block, ttc_gfx_stdout_print_string);

    PoolPrintJobs = ttc_memory_pool_create(sizeof(printJob_t), 20);
    ttc_list_init(&BlockList, "BlockList");

    switch (DMA_Algorithm) { // initialize dynamic data structures
    case dma_FCTMA:
        Pool100 = ttc_memory_pool_create( Block_Size, 160 );
        break;
    default:
        break;
    }

    static u8_t Temp3[20];
    Base_t SleepMax;
    SleepMax = 10;
    for (u8_t Index = 0; Index < Amount_Consumer; Index++) {
        snprintf(Temp3, 20, "Consumer%02i", Index);
        EndOfConsumerTasks->SleepTime_Max = SleepMax;
        ttc_task_create(taskConsumer, (char*) Temp3, 256, (void*) EndOfConsumerTasks++, 0, NULL);
        ttc_task_check_stack();
    }

    SleepMax = 15;
    for (u8_t Index = 0; Index < Amount_Producer; Index++) {
        snprintf(Temp3, 20, "Producer%02i", Index);
        EndOfProducerTasks->SleepTime_Max = SleepMax;
        ttc_task_create(taskProducer, (char*) Temp3, 256, (void*) EndOfProducerTasks++, 0, NULL);
        ttc_task_check_stack();
    }

    ttc_task_create(taskRegression_MemoryPools, "MemoryPools", 256, NULL, 1, NULL);
}
void taskRegression_MemoryPools(void* Arg) {
    (void) Arg;

    switch (DMA_Algorithm) {
    case dma_FCTMA:
        printAt(0,0, "Test: FCTMA", 0);
        break;
    case dma_FreeRTOS:
        printAt(0,0, "Test: FreeRTOS", 0);
        break;
    case dma_SDMA:
        printAt(0,0, "Test: SDMA", 0);
        break;
    default:
        Assert_Halt_EC(ec_NotImplemented);
        break;
    }

    printAt(2, DisplayConfig->TextRows - 1, "used ", 0);
    AmountHeapFreeAtStart = ttc_memory_get_free_size();

    Base_t P_BlocksProcessed, P_SleepTime_Total;
    Base_t C_BlocksProcessed, C_SleepTime_Total;
    TaskInfo_t* TaskInfo = NULL;
    Base_t SleepTime_Average;

    while (1) {
        u8_t Row = 1;

        P_BlocksProcessed   = 1;
        P_SleepTime_Total   = 0;
        TaskInfo = ProducerTasks;
        while (TaskInfo < EndOfProducerTasks) { // collect statistic data of all producer tasks
            P_BlocksProcessed += TaskInfo->BlocksProcessed;
            P_SleepTime_Total += TaskInfo->SleepTime_Total;
            TaskInfo++;
        }

        C_BlocksProcessed   = 1;
        C_SleepTime_Total   = 0;
        TaskInfo = ConsumerTasks;
        while (TaskInfo < EndOfConsumerTasks) { // collect statistic data of all consumer tasks
            C_BlocksProcessed += TaskInfo->BlocksProcessed;
            C_SleepTime_Total += TaskInfo->SleepTime_Total;
            TaskInfo++;
        }

        printAt(0, Row, "P(%2i)", Amount_Producer);
        printAt(6, Row, "#%07i",  P_BlocksProcessed);
        Row++;
        SleepTime_Average = DIV(P_SleepTime_Total, P_BlocksProcessed);
        printAt( 0, Row++, "SleepAvg %3i ", SleepTime_Average);
        printAt(0, Row, "C(%2i)", Amount_Consumer);
        printAt(6, Row, "#%07i",  C_BlocksProcessed);
        Row++;
        SleepTime_Average = DIV(C_SleepTime_Total, C_BlocksProcessed);
        printAt( 0, Row++, "SleepAvg %3i ", SleepTime_Average);

        if (C_BlocksProcessed > 10000)
            ttc_testpoint();

        if (DMA_Algorithm == dma_FCTMA) { // display extra pool-statistics
#if TTC_MEMORY_POOL_STATISTICS == 1
            printAt( 0, Row++, "Alloc    %4i", Pool100->Statistics.AmountAllocated);
            printAt( 0, Row++, "MaxFree  %4i", Pool100->Statistics.AmountFreeMaximum);
            printAt( 0, Row++, "Available%4i", ttc_semaphore_available(&(Pool100->BlocksAvailable)));
            printAt( 0, Row++, "Free     %4i", Pool100->Statistics.AmountFreeCurrently);
#endif
        }
        printAt( 0, Row++, "Total  %6i", ttc_memory_Bytes_Allocated_Total);
        printAt( 0, Row++, "Usable %6i", ttc_memory_Bytes_Allocated_Usable);
        printAt( 0, Row++, "Alloc# %3i", ttc_memory_Amount_Blocks_Allocated);


        if (0) printAt(8, DisplayConfig->TextRows - 1, "%iB", AmountHeapFreeAtStart - ttc_memory_get_free_size());
        static u8_t Count = 0;
        if (Count++ & 1)
            printAt(0, DisplayConfig->TextRows - 1, "+", 0);
        else
            printAt(0, DisplayConfig->TextRows - 1, "-", 0);
        mSleep(10000);
    }
}
void taskProducer(void* Arg) {
    TaskInfo_t* MyTaskInfo = (TaskInfo_t*) Arg;
    //mSleep(700); // Will cause hard-fault!

    void* Block = NULL;
    ttc_gpio_init(PIN_PC6, tgm_output_push_pull); //D

    while (1) {
#ifdef TTC_REGRESSION
        ttc_task_check_stack();
#endif

        ttc_gpio_set(PIN_PC6); //D
        for (u8_t I = Amount_BlocksPerTurn; I > 0; I--) {
            Block = mallocBlock();

            if (Block) {
                MyTaskInfo->BlocksProcessed++;
                unsigned char* Data = (unsigned char*) Block;
                (void) Data;

                push2Consumer(Block);
            }
            else {
                ttc_task_yield();
            }
        }
        ttc_gpio_clr(PIN_PC6); //D

        if (MyTaskInfo->BlocksProcessed)
            MyTaskInfo->SleepTime_Average = DIV(MyTaskInfo->SleepTime_Total, MyTaskInfo->BlocksProcessed / Amount_BlocksPerTurn);

        if (ttc_memory_Bytes_Allocated_Total > 16384) { // force allocation until certain amount of memory has been allocated
            // simulate random time required to fill Data[]
            Base_t SleepTime = 1 + random(MyTaskInfo->SleepTime_Max);
            MyTaskInfo->SleepTime_Total += SleepTime;
            mSleep(SleepTime);
        }
    }
}
void taskConsumer(void* Arg) {
    TaskInfo_t* MyTaskInfo = (TaskInfo_t*) Arg;
    //mSleep(500); // Will cause hard-fault!

    while (1) {
        ttc_task_check_stack(); //D

        // simulate random processing time of Data[]
        Base_t SleepTime = 1 + random(MyTaskInfo->SleepTime_Max);
        MyTaskInfo->SleepTime_Total += SleepTime;
        mSleep(SleepTime);

        for (u8_t I = Amount_BlocksPerTurn; I > 0; I--) {
            void *Block = popFromProducer();
            if (Block) {
//                if (DMA_Algorithm == dma_FCTMA) {
//                    ttc_list_item_t* Item = (ttc_list_item_t*) Block;
//                    Block = (void*) ttc_memory_pool_from_list_item(Item);
//                    Assert(ttc_memory_is_writable(Block), ec_InvalidImplementation);
//                }
                unsigned char* Data = (unsigned char*) Block;
                (void) Data;

                releaseBlock(Block);
                MyTaskInfo->BlocksProcessed++;
            }
            else
                ttc_task_yield();
        }
        if (MyTaskInfo->BlocksProcessed)
            MyTaskInfo->SleepTime_Average = DIV(MyTaskInfo->SleepTime_Total, MyTaskInfo->BlocksProcessed / Amount_BlocksPerTurn);
    }
}
void printAt(u8_t X, u8_t Y, const char* Format, u32_t Value) {
    static TaskHandle_t* MyTask = NULL;
    if (MyTask) {
        Assert(MyTask == ttc_task_current_handle(), ec_InvalidImplementation); // this function may only be used from one task
    }
    else {
        MyTask = ttc_task_current_handle(); // called first time: store owner task
        Assert(ttc_memory_is_writable(MyTask), ec_InvalidImplementation);
    }
    static u8_t Temp1[100]; // there can be only one taskPrint() instance
    ttc_gfx_text_cursor_set(X, Y);
    ttc_string_snprintf(Temp1, 100, Format, Value);
    ttc_gfx_print_solid( (char*) Temp1, 100);
}

//} Function definitions
